"""Module to determine the pywikibot version (tag, revision and date)."""
#
# (C) Pywikibot team, 2007-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import datetime
import json
import os
import pathlib
import socket
import subprocess
import sys
import sysconfig
import time
from contextlib import suppress
from importlib import import_module
from pathlib import Path
from warnings import warn

import pywikibot
from pywikibot import config
from pywikibot.backports import cache
from pywikibot.comms.http import fetch
from pywikibot.exceptions import VersionParseError
from pywikibot.tools import deprecated


def _get_program_dir() -> str:
    return os.path.normpath(os.path.split(os.path.dirname(__file__))[0])


@deprecated(since='9.0.0')
def get_toolforge_hostname() -> str | None:
    """Get hostname of the current Toolforge host.

    .. versionadded:: 3.0
    .. deprecated:: 9.0

    :return: The hostname of the currently running host,
             if it is in Wikimedia Toolforge; otherwise return None.
    """
    if socket.getfqdn().endswith('.tools.eqiad1.wikimedia.cloud'):
        return socket.gethostname()
    return None


def getversion(online: bool = True) -> str:
    """Return a pywikibot version string.

    :param online: Include information obtained online
    """
    branches = {
        'master': 'branches/master',
        'stable': 'branches/stable',
    }
    data = getversiondict()
    data['cmp_ver'] = 'n/a'
    local_hsh = data.get('hsh', '')
    hsh = {}

    if online:
        if not local_hsh:
            data['cmp_ver'] = 'UNKNOWN'
        else:
            for branch, path in branches.items():
                with suppress(VersionParseError):
                    hsh[getversion_onlinerepo(path)] = branch
            if hsh:
                data['cmp_ver'] = hsh.get(local_hsh, 'OUTDATED')

    data['hsh'] = local_hsh[:7]  # make short hash from full hash
    return '{tag} ({hsh}, {rev}, {date}, {cmp_ver})'.format_map(data)


@cache
def getversiondict() -> dict[str, str]:
    """Get version info for the package.

    :return: Return a dict with the following keys:

        - tag (name for the repository),
        - rev (current revision identifier),
        - date (date of current revision),
        - hash (git hash for the current revision)
    """
    _program_dir = _get_program_dir()
    exceptions = {}

    for vcs_func in (getversion_git,
                     getversion_nightly,
                     getversion_package):
        try:
            tag, rev, date, hsh = vcs_func(_program_dir)
        except Exception as e:
            exceptions[vcs_func] = vcs_func.__name__, e
        else:
            break
    else:  # pragma: no cover
        # nothing worked; version unknown (but suppress exceptions)
        # the value is most likely '$Id' + '$', it means that
        # pywikibot was imported without using version control at all.
        tag, rev, date, hsh = (
            '', '-1 (unknown)', '0 (unknown)', '(unknown)')
        warn(f'Unable to detect version; exceptions raised:\n{exceptions!r}',
             UserWarning, stacklevel=2)
        exceptions = None

    # Git and SVN can silently fail, as it may be a nightly.
    if exceptions:  # pragma: no cover
        pywikibot.debug(f'version algorithm exceptions:\n{exceptions!r}')

    if isinstance(date, str):
        datestring = date
    elif isinstance(date, time.struct_time):
        datestring = time.strftime('%Y/%m/%d, %H:%M:%S', date)
    else:  # pragma: no cover
        warn('Unable to detect package date', UserWarning, stacklevel=2)
        datestring = '-2 (unknown)'

    return {'tag': tag, 'rev': rev, 'date': datestring, 'hsh': hsh}


def getversion_git(path=None):
    """Get version info for a Git clone.

    :param path: directory of the Git checkout
    :return:
        - tag (name for the repository),
        - rev (current revision identifier),
        - date (date of current revision),
        - hash (git hash for the current revision)
    :rtype: ``tuple`` of three ``str`` and a ``time.struct_time``
    """
    _program_dir = path or _get_program_dir()
    cmd = 'git'
    try:
        subprocess.Popen([cmd], stdout=subprocess.PIPE).communicate()
    except OSError:
        # some Windows git versions provide git.cmd instead of git.exe
        cmd = 'git.cmd'

    with open(os.path.join(_program_dir, '.git/config')) as f:
        tag = f.read()
    # Try 'origin' and then 'gerrit' as remote name; bail if can't find either.
    remote_pos = tag.find('[remote "origin"]')
    if remote_pos == -1:
        remote_pos = tag.find('[remote "gerrit"]')
    if remote_pos == -1:
        tag = '?'
    else:
        s = tag.find('url = ', remote_pos)
        e = tag.find('\n', s)
        tag = tag[(s + 6):e]
        t = tag.strip().split('/')
        tag = f"[{t[0][:-1]}] {'-'.join(t[3:])}"
    dp = subprocess.Popen([cmd, '--no-pager',
                           'log', '-1',
                           '--pretty=format:"%ad|%an|%h|%H|%d"',
                           '--abbrev-commit',
                           '--date=iso'],
                          cwd=_program_dir,
                          stdout=subprocess.PIPE)
    info, _ = dp.communicate()
    info = info.decode(config.console_encoding).split('|')
    date = info[0][:-6]
    date = time.strptime(date.strip('"'), '%Y-%m-%d %H:%M:%S')
    dp = subprocess.Popen([cmd, 'rev-list', 'HEAD'],
                          cwd=_program_dir,
                          stdout=subprocess.PIPE)
    rev, stderr = dp.communicate()
    rev = f'g{len(rev.splitlines())}'
    hsh = info[3]  # also stored in '.git/refs/heads/master'
    if (not date or not tag or not rev) and not path:
        raise VersionParseError
    return (tag, rev, date, hsh)


def getversion_nightly(path: str | Path | None = None):
    """Get version info for a nightly release.

    .. hint::
       the version information of the nightly dump is stored in the
       ``version`` file within the ``pywikibot`` folder.

    :param path: directory of the uncompressed nightly.
    :return:
        - tag (name for the repository),
        - rev (current revision identifier),
        - date (date of current revision),
        - hash (git hash for the current revision)
    :rtype: ``tuple`` of three ``str`` and a ``time.struct_time``
    """
    file = Path(path or _get_program_dir())
    if not path:
        file /= 'pywikibot'  # pragma: no cover
    file /= 'version'

    with file.open() as data:
        (tag, rev, date, hsh) = data.read().splitlines()

    date = time.strptime(date[:19], '%Y-%m-%dT%H:%M:%S')

    if not date or not tag or not rev:
        raise VersionParseError  # pragma: no cover
    return (tag, rev, date, hsh)


def getversion_package(path=None) -> tuple[str, str, str, str]:
    """Get version info for an installed package.

    :param path: Unused argument
    :return: Return a tuple with the following items:

        - tag: 'pywikibot/__init__.py'
        - rev: '-1 (unknown)'
        - date (date the package was installed locally),
        - hash (git hash for the current revision of 'pywikibot/__init__.py')
    """
    hsh = ''
    date = get_module_mtime(pywikibot).timetuple()

    tag = 'pywikibot/__init__.py'
    rev = '-1 (unknown)'

    return (tag, rev, date, hsh)


def getversion_onlinerepo(path: str = 'branches/master') -> str:
    """Retrieve current framework git hash from Gerrit."""
    # Gerrit API responses include )]}' at the beginning,
    # make sure to strip it out
    buf = fetch(
        'https://gerrit.wikimedia.org/r/projects/pywikibot%2Fcore/' + path,
        headers={'user-agent': '{pwb}'}).text[4:]
    try:
        return json.loads(buf)['revision']
    except Exception as e:  # pragma: no cover
        raise VersionParseError(f'{e!r} while parsing {buf!r}')


def get_module_filename(module) -> str | None:
    """Retrieve filename from an imported pywikibot module.

    It uses the __file__ attribute of the module. If it's file extension
    ends with py and another character the last character is discarded
    when the py file exist.

    :param module: The module instance.
    :type module: module
    :return: The filename if it's a pywikibot module otherwise None.
    """
    if hasattr(module, '__file__'):
        filename = module.__file__
        if not filename or not os.path.exists(filename):
            return None

        program_dir = _get_program_dir()
        if filename.startswith(program_dir):
            return filename
    return None


def get_module_mtime(module):
    """Retrieve the modification time from an imported module.

    :param module: The module instance.
    :type module: module
    :return: The modification time if it's a pywikibot module otherwise
        None.
    :rtype: datetime or None
    """
    filename = get_module_filename(module)
    if filename:
        return datetime.datetime.fromtimestamp(os.stat(filename).st_mtime)
    return None


def package_versions(
    modules: list[str] | None = None,
    builtins: bool | None = False,
    standard_lib: bool | None = None,
):
    """Retrieve package version information.

    When builtins or standard_lib are None, they will be included only
    if a version was found in the package.

    :param modules: Modules to inspect
    :param builtins: Include builtins
    :param standard_lib: Include standard library packages
    """
    if not modules:
        modules = sys.modules.keys()

    std_lib_dir = pathlib.Path(sysconfig.get_paths()['stdlib'])

    root_packages = {key.split('.')[0] for key in modules}

    builtin_packages = {name.split('.')[0] for name in root_packages
                        if name in sys.builtin_module_names
                        or '_' + name in sys.builtin_module_names}

    # Improve performance by removing builtins from the list if possible.
    if builtins is False:
        root_packages -= builtin_packages

    std_lib_packages = []

    paths = {}
    data = {}

    for name in root_packages:
        try:
            package = import_module(name)
        except ImportError as e:
            data[name] = {'name': name, 'err': e}
            continue

        info = {'package': package, 'name': name}

        if name in builtin_packages:
            info['type'] = 'builtins'

        if '__file__' in package.__dict__:
            # Determine if this file part is of the standard library.
            # possible Namespace package
            if not hasattr(package, '__file__') or package.__file__ is None:
                _file = None
                _path = pathlib.Path(package.__path__[0])
            else:
                _file = pathlib.Path(package.__file__)
                _path = _file.parent
            if _path == std_lib_dir:
                std_lib_packages.append(name)
                if standard_lib is False:
                    continue
                info['type'] = 'standard library'

            # Strip '__init__.py' from the filename.
            if (not hasattr(package, '__file__')
                    or package.__file__ is None
                    or _file.name == '__init__.py'):
                path = _path
            else:
                path = _file

            info['path'] = path
            assert path not in paths, (
                f'Path {path} of the package {name} is in defined paths as '
                f'{paths[path]}'
            )

            paths[path] = name

        if '__version__' in package.__dict__:
            info['ver'] = package.__version__
        elif name.startswith('unicodedata'):
            info['ver'] = package.unidata_version

        # If builtins or standard_lib is None,
        # only include package if a version was found.
        if builtins is None and name in builtin_packages \
           or standard_lib is None and name in std_lib_packages:
            if 'ver' in info:
                data[name] = info
            else:
                # Remove the entry from paths, so it isn't processed below
                del paths[info['path']]
        else:
            data[name] = info

    # Remove any pywikibot sub-modules which were loaded as a package.
    # e.g. 'wikipedia_family.py' is loaded as 'wikipedia'
    _program_dir = _get_program_dir()
    dir_parts = pathlib.Path(_program_dir).parts
    length = len(dir_parts)
    for path, name in paths.items():
        lib_parts = path.parts
        if dir_parts != lib_parts[:length]:
            continue
        if lib_parts[length] != '.tox':
            del data[name]

    return data



"""Objects representing API upload to MediaWiki sites."""
#
# (C) Pywikibot team, 2009-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import mimetypes
import os
from collections.abc import Iterable
from warnings import warn

import pywikibot
from pywikibot.exceptions import APIError, Error, UploadError
from pywikibot.tools import compute_file_hash


__all__ = ('Uploader', )


class Uploader:

    """Uploader class to upload a file to the wiki.

    .. versionadded:: 7.1

    :param site: The current site to work on
    :param filepage: a FilePage object from which the wiki-name of the
        file will be obtained.
    :param source_filename: path to the file to be uploaded
    :param source_url: URL of the file to be uploaded
    :param comment: Edit summary; if this is not provided, then
        filepage.text will be used. An empty summary is not permitted.
        This may also serve as the initial page text (see below).
    :param text: Initial page text; if this is not set, then
        filepage.text will be used, or comment.
    :param watch: If true, add filepage to the bot user's watchlist
    :param chunk_size: The chunk size in bytes for chunked uploading
        (see :api:`Upload#Chunked_uploading`). It will only upload in
        chunks, if the chunk size is positive but lower than the file
        size.
    :param asynchronous: Make potentially large file operations
        asynchronous on the server side when possible.
    :param ignore_warnings: It may be a static boolean, a callable
        returning a boolean or an iterable. The callable gets a list of
        UploadError instances and the iterable should contain the warning
        codes for which an equivalent callable would return True if all
        UploadError codes are in that list. If the result is False it'll
        not continue uploading the file and otherwise disable any warning
        and reattempt to upload the file.

        .. note:: If report_success is True or None it'll raise an
           UploadError exception if the static boolean is False.
    :type ignore_warnings: bool or callable or iterable of str
    :param report_success: If the upload was successful it'll print a
        success message and if ignore_warnings is set to False it'll
        raise an UploadError if a warning occurred. If it's None
        (default) it'll be True if ignore_warnings is a bool and False
        otherwise. If it's True or None ignore_warnings must be a bool.
    """

    upload_warnings = {
        # map API warning codes to user error messages
        # {msg} will be replaced by message string from API response
        'duplicate-archive':
            'The file is a duplicate of a deleted file {msg}.',
        'was-deleted': 'The file {msg} was previously deleted.',
        'empty-file': 'File {msg} is empty.',
        'exists': 'File {msg} already exists.',
        'duplicate': 'Uploaded file is a duplicate of {msg}.',
        'badfilename': 'Target filename is invalid.',
        'filetype-unwanted-type': 'File {msg} type is unwanted type.',
        'exists-normalized':
            'File exists with different extension as {msg!r}.',
        'bad-prefix': 'Target filename has a bad prefix {msg}.',
        'page-exists':
            'Target filename exists but with a different file {msg}.',

        # API-returned message string will be timestamps, not much use here
        'no-change': 'The upload is an exact duplicate of the current version '
                     'of this file.',
        'duplicate-version': 'The upload is an exact duplicate of older '
                             'version(s) of this file.',
    }

    def __init__(self,
                 site: pywikibot.site.APISite,
                 filepage: pywikibot.FilePage,
                 *,
                 source_filename: str | None = None,
                 source_url: str | None = None,
                 comment: str | None = None,
                 text: str | None = None,
                 watch: bool = False,
                 chunk_size: int = 0,
                 asynchronous: bool = False,
                 ignore_warnings=False,
                 report_success: bool | None = None) -> None:
        """Initializer."""
        self.site = site
        self.filepage = filepage
        self.comment = comment
        self.text = text
        self.watch = watch
        self.ignore_warnings = ignore_warnings
        self.chunk_size = chunk_size
        self.asynchronous = asynchronous
        self.report_success = report_success

        if source_filename and source_url:
            raise ValueError('APISite.upload: must provide either '
                             'source_filename or source_url, not both.')

        self.filename = source_filename
        self.url = source_url

    def upload(self) -> bool:
        """Check for required parameters to upload and run the job.

        :return: Whether the upload was successful.
        """
        if self.comment is None:
            self.comment = self.filepage.text
        if not self.comment:
            raise ValueError('APISite.upload: cannot upload file without '
                             'a summary/description.')

        if self.text is None:
            self.text = self.filepage.text
        if not self.text:
            self.text = self.comment

        return self._upload(self.ignore_warnings, self.report_success)

    @classmethod
    def create_warnings_list(cls, response, file_key):
        """Create a list of upload errors."""
        return [UploadError(warning,
                            cls.upload_warnings.get(warning, '{msg}')
                            .format(msg=data),
                            file_key,
                            response['offset'])
                for warning, data in response['warnings'].items()]

    def _upload(self, ignore_warnings, report_success,
                file_key=None, offset=0) -> bool:
        """Recursive Upload method.

        :param file_key: Reuses an already uploaded file using the
            filekey. If None (default) it will upload the file.
        :param offset: When file_key is not None this can be an integer
            to continue a previously canceled chunked upload. If False
            it treats that as a finished upload. If True it requests the
            stash info from the server to determine the offset. By
            default starts at 0.
        :return: Whether the upload was successful.
        """
        # An offset != 0 doesn't make sense without a file key
        assert offset == 0 or file_key is not None

        if report_success is None:
            report_success = isinstance(ignore_warnings, bool)
        if report_success is True and not isinstance(ignore_warnings, bool):
            raise ValueError('report_success may only be set to True when '
                             'ignore_warnings is a boolean')
        if isinstance(ignore_warnings, Iterable):
            ignored_warnings = ignore_warnings

            def ignore_warnings(warnings):
                return all(w.code in ignored_warnings for w in warnings)

        ignore_all_warnings = not callable(ignore_warnings) and ignore_warnings

        token = self.site.tokens['csrf']
        result = None
        file_page_title = self.filepage.title(
            with_ns=False,
            with_section=False,
        )
        file_size = None

        # make sure file actually exists
        if self.filename:
            if os.path.isfile(self.filename):
                file_size = os.path.getsize(self.filename)
            elif offset is not False:
                raise ValueError(f"File '{self.filename}' does not exist.")

        # Verify the stash when a file key and offset is given:
        # requests the SHA1 and file size uploaded and compares it to
        # the local file. Also verify that offset is matching the
        # file size if the offset is an int. If offset is False if
        # verifies that the file size match with the local file.
        verify_stash = False
        if self.filename and file_key:
            assert offset is False or file_size is not None
            verify_stash = True
            if (offset is not False and offset is not True
                    and offset > file_size):
                raise ValueError(
                    f'For the file key "{file_key}" the offset was set to '
                    f'{offset} while the file is only {file_size} bytes large.'
                )

        if verify_stash or offset is True:
            if not file_key:
                raise ValueError('Without a file key it cannot request the '
                                 'stash information')
            if not self.filename:
                raise ValueError('Can request stash information only when '
                                 'using a file name.')
            props = ['size']
            if verify_stash:
                props.append('sha1')
            stash_info = self.site.stash_info(file_key, props)
            if offset is True:
                offset = stash_info['size']
            elif offset is False:
                if file_size != stash_info['size']:
                    raise ValueError(
                        'For the file key "{}" the server reported a size '
                        '{} while the file size is {}'
                        .format(file_key, stash_info['size'], file_size))
            elif offset is not False and offset != stash_info['size']:
                raise ValueError(
                    'For the file key "{}" the server reported a size {} '
                    'while the offset was {}'
                    .format(file_key, stash_info['size'], offset))

            if verify_stash:
                # The SHA1 was also requested so calculate and compare it
                assert 'sha1' in stash_info, \
                    f'sha1 not in stash info: {stash_info}'
                sha1 = compute_file_hash(self.filename, bytes_to_read=offset)
                if sha1 != stash_info['sha1']:
                    raise ValueError(
                        'The SHA1 of {} bytes of the stashed "{}" is {} '
                        'while the local file is {}'
                        .format(offset, file_key, stash_info['sha1'], sha1))

        assert offset is not True
        if file_key and file_size is None:
            assert offset is False

        data = {}
        if file_key and offset is False or offset == file_size:
            pywikibot.log(
                f'Reused already upload file using filekey "{file_key}"')
            # TODO: Use sessionkey instead of filekey if necessary
            final_request = self.site._request(
                parameters={
                    'action': 'upload',
                    'token': token,
                    'filename': file_page_title,
                    'comment': self.comment,
                    'text': self.text,
                    'async': self.asynchronous,
                    'filekey': file_key
                })

        elif self.filename:
            # TODO: Dummy value to allow also Unicode names, see bug T75661
            mime_filename = 'FAKE-NAME'
            # upload local file
            throttle = True
            filesize = os.path.getsize(self.filename)
            chunked_upload = 0 < self.chunk_size < filesize
            with open(self.filename, 'rb') as f:
                final_request = self.site._request(
                    throttle=throttle, parameters={
                        'action': 'upload', 'token': token, 'text': self.text,
                        'filename': file_page_title, 'comment': self.comment})
                if chunked_upload:
                    if offset > 0:
                        pywikibot.log(f'Continuing upload from byte {offset}')
                    poll = False
                    while True:

                        if poll:
                            # run a poll; not possible in first iteration
                            assert file_key
                            req = self.site.simple_request(
                                action='upload',
                                token=token,
                                filekey=file_key,
                                checkstatus=True)
                        else:
                            f.seek(offset)
                            chunk = f.read(self.chunk_size)
                            # workaround (hack) for T132676
                            # append another '\r' so that one is the payload
                            # and the second is used for newline when mangled
                            # by email package.
                            if (len(chunk) < self.chunk_size
                                    or (offset + len(chunk)) == filesize
                                    and chunk[-1] == b'\r'[0]):
                                chunk += b'\r'

                            mime_params = {
                                'chunk': (chunk,
                                          ('application', 'octet-stream'),
                                          {'filename': mime_filename})
                            }
                            req = self.site._request(
                                throttle=throttle,
                                mime=mime_params,
                                parameters={
                                    'action': 'upload',
                                    'token': token,
                                    'stash': True,
                                    'filesize': filesize,
                                    'offset': offset,
                                    'filename': file_page_title,
                                    'async': self.asynchronous,
                                    'ignorewarnings': ignore_all_warnings})

                            if file_key:
                                req['filekey'] = file_key

                        try:
                            data = req.submit()['upload']
                        except APIError as error:
                            # TODO: catch and process foreseeable errors
                            if error.code == 'stashfailed' \
                                    and 'offset' in error.other:
                                # TODO: Ask MediaWiki to change this
                                # ambiguous error code.

                                new_offset = int(error.other['offset'])
                                # If the offset returned from the server
                                # (the offset it expects now) is equal to
                                # the offset we sent it, there must be
                                # something else that prevented the upload,
                                # instead of simple offset mismatch. This
                                # also prevents infinite loops when we
                                # upload the same chunk again and again,
                                # every time ApiError.
                                if offset != new_offset:
                                    pywikibot.log(
                                        f'Old offset: {offset}; Returned '
                                        f'offset: {new_offset}; Chunk size: '
                                        f'{len(chunk)}'
                                    )
                                    pywikibot.warning('Attempting to correct '
                                                      'automatically from '
                                                      'offset mismatch error.')
                                    offset = new_offset
                                    continue
                            raise
                        if 'nochange' in data:  # in simulation mode
                            break

                        # Polls may not contain file key in response
                        file_key = data.get('filekey', file_key)
                        if data['result'] == 'Warning':
                            assert ('warnings' in data
                                    and not ignore_all_warnings)
                            if callable(ignore_warnings):
                                restart = False
                                if 'offset' not in data:
                                    # This is a result of a warning in the
                                    # first chunk. The chunk is not actually
                                    # stashed so upload must be restarted if
                                    # the warning is allowed.
                                    # T112416 and T112405#1637544
                                    restart = True
                                    data['offset'] = True
                                if ignore_warnings(self.create_warnings_list(
                                        data, file_key)):
                                    # Future warnings of this run
                                    # can be ignored
                                    if restart:
                                        return self._upload(
                                            ignore_warnings=True,
                                            report_success=False
                                        )

                                    ignore_warnings = True
                                    ignore_all_warnings = True
                                    offset = data['offset']
                                    continue
                                return False
                            result = data
                            result.setdefault('offset', 0)
                            break

                        if data['result'] == 'Continue':
                            throttle = False
                            if 'offset' in data:
                                new_offset = int(data['offset'])
                                if offset + len(chunk) != new_offset:
                                    pywikibot.log(
                                        f'Old offset: {offset}; Returned '
                                        f'offset: {new_offset}; Chunk size: '
                                        f'{len(chunk)}'
                                    )
                                    pywikibot.warning('Unexpected offset.')
                                offset = new_offset
                            else:
                                pywikibot.warning('Offset was not supplied.')
                                offset += len(chunk)
                        elif data['result'] == 'Poll':
                            poll = True
                            pywikibot.log('Waiting for server to '
                                          'assemble chunks.')
                        elif data['result'] == 'Success':  # finished
                            pywikibot.log('Finished uploading last chunk.')
                            final_request['filekey'] = file_key
                            final_request['async'] = self.asynchronous
                            break
                        else:
                            raise Error('Unrecognized result: {result}'
                                        .format_map(data))

                elif file_key:
                    final_request['filekey'] = file_key
                else:
                    file_contents = f.read()
                    filetype = (mimetypes.guess_type(self.filename)[0]
                                or 'application/octet-stream')
                    final_request.mime = {
                        'file': (file_contents, filetype.split('/'),
                                 {'filename': mime_filename})
                    }
        else:
            # upload by URL
            if not self.site.has_right('upload_by_url'):
                raise Error(f"User '{self.site.user()}' is not authorized to "
                            f'upload by URL on site {self}.')
            final_request = self.site.simple_request(
                action='upload', filename=file_page_title, url=self.url,
                comment=self.comment, text=self.text, token=token)

        return self.submit(final_request, result, data.get('result'),
                           ignore_warnings, ignore_all_warnings,
                           report_success, file_key)

    def submit(self, request, result, data_result: str | None,
               ignore_warnings, ignore_all_warnings, report_success,
               file_key) -> bool:
        """Submit request and return whether upload was successful."""
        # some warning keys have been changed
        warning_keys = {
            'nochange': 'no-change',
            'duplicateversions': 'duplicate-version',
            'emptyfile': 'empty-file',
        }

        token = request['token']
        while True:
            if not result:
                request['watch'] = self.watch
                request['ignorewarnings'] = ignore_all_warnings
                result = request.submit()['upload']
                pywikibot.debug(result)

            if 'result' not in result:
                raise Error(f'Upload: unrecognized response: {result}')

            if result['result'] == 'Warning':
                assert 'warnings' in result and not ignore_all_warnings

                if self.filename:
                    if 'filekey' in result:
                        file_key = result['filekey']
                    elif 'sessionkey' in result:
                        # TODO: Probably needs to be reflected in the API call
                        # above
                        file_key = result['sessionkey']
                        pywikibot.warning(
                            'Using sessionkey instead of filekey.')
                    else:
                        file_key = None
                        pywikibot.warning('No filekey defined.')
                else:
                    file_key = None

                if not report_success:
                    result.setdefault('offset', bool(self.filename))
                    offset = result['offset'] if self.filename else False
                    if ignore_warnings(self.create_warnings_list(result,
                                                                 file_key)):
                        return self._upload(ignore_warnings=True,
                                            report_success=False,
                                            file_key=file_key,
                                            offset=offset)
                    return False

                if len(result['warnings']) > 1:
                    warn('The upload returned {} warnings: {}'
                         .format(len(result['warnings']),
                                 ', '.join(result['warnings'])),
                         UserWarning, 3)
                warning = list(result['warnings'].keys())[0]
                message = result['warnings'][warning]
                warning = warning_keys.get(warning, warning)
                raise UploadError(warning,
                                  self.upload_warnings[warning]
                                  .format(msg=message),
                                  file_key=file_key,
                                  offset=result.get('offset', False))

            if result['result'] == 'Poll':
                # Polling is meaningless without a file key
                assert file_key
                pywikibot.log('Waiting for upload to be published.')
                result = None
                request = self.site.simple_request(
                    action='upload',
                    token=token,
                    filekey=file_key,
                    checkstatus=True)
                continue

            if result['result'] == 'Success':
                if report_success:
                    pywikibot.info('Upload successful.')
                # If we receive a nochange, that would mean we're in simulation
                # mode, don't attempt to access imageinfo
                if 'nochange' not in result:
                    self.filepage._load_file_revisions([result['imageinfo']])
                return True

            raise Error(
                f"Unrecognized result: {data_result or result['result']}")



"""Iterator functions.

.. note:: ``pairwise()`` function introduced in Python 3.10 is backported
   in :mod:`backports`
"""
#
# (C) Pywikibot team, 2008-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import collections
import heapq
import itertools
from contextlib import suppress
from typing import Any

from pywikibot.backports import (
    Callable,
    Generator,
    Iterable,
    Iterator,
    batched,
)
from pywikibot.logging import debug
from pywikibot.tools import deprecated


__all__ = (
    'filter_unique',
    'intersect_generators',
    'islice_with_ellipsis',
    'itergroup',
    'roundrobin_generators',
    'union_generators',
)


@deprecated('backports.batched()', since='8.2.0')
def itergroup(iterable,
              size: int,
              strict: bool = False) -> Generator[list[Any], None, None]:
    """Make an iterator that returns lists of (up to) size items from iterable.

    Example:

    >>> i = itergroup(range(25), 10)
    >>> print(next(i))
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    >>> print(next(i))
    [10, 11, 12, 13, 14, 15, 16, 17, 18, 19]
    >>> print(next(i))
    [20, 21, 22, 23, 24]
    >>> print(next(i))
    Traceback (most recent call last):
     ...
    StopIteration

    .. versionadded:: 7.6
       The *strict* parameter.
    .. deprecated:: 8.2
       Use :func:`backports.batched` instead.

    :param size: How many items of the iterable to get in one chunk
    :param strict: If True, raise a ValueError if length of iterable is
        not divisible by `size`.
    :raises ValueError: iterable is not divisible by size
    """
    for group in batched(iterable, size, strict=strict):
        yield list(group)


def islice_with_ellipsis(iterable, *args, marker: str = '…'):
    """Generator which yields the first n elements of the iterable.

    If more elements are available and marker is True, it returns an extra
    string marker as continuation mark.

    Function takes the
    and the additional keyword marker.

    :param iterable: the iterable to work on
    :type iterable: iterable
    :param args: same args as:
        - ``itertools.islice(iterable, stop)``
        - ``itertools.islice(iterable, start, stop[, step])``
    :param marker: element to yield if iterable still contains elements
        after showing the required number. Default value: '…'
    """
    s = slice(*args)
    _iterable = iter(iterable)
    yield from itertools.islice(_iterable, *args)
    if marker and s.stop is not None:
        with suppress(StopIteration):
            next(_iterable)
            yield marker


def union_generators(*iterables: Iterable[Any],
                     key: Callable[[Any], Any] | None = None,
                     reverse: bool = False) -> Iterator[Any]:
    """Generator of union of sorted iterables.

    Yield all items from the input iterables in sorted order, removing
    duplicates. The input iterables must already be sorted according to
    the same *key* and direction. For descending direction, *reverse*
    must be ``True``. The generator will yield each element only once,
    even if it appears in multiple iterables. This behaves similarly to:

        sorted(set(itertools.chain(*iterables)), key=key, reverse=reverse)

    but is memory-efficient since it processes items lazily.

    Sample:

    >>> list(union_generators([1, 2, 3, 4], [3, 4, 5], [2, 6]))
    [1, 2, 3, 4, 5, 6]
    >>> list(union_generators([4, 3, 2, 1], [5, 4, 3], [6, 2], reverse=True))
    [6, 5, 4, 3, 2, 1]

    .. versionadded:: 10.6

    .. note::
       All input iterables must be sorted consistently. *reverse* must
       be set to ``True`` only if the iterables are sorted in descending
       order. For simple concatenation without duplicate removal, use
       :pylib:`itertools.chain<itertools#itertools.chain>` instead.

    :param iterables: Sorted iterables to merge.
    :param key: Optional key function to compare elements. If ``None``,
        items are compared directly.
    :param reverse: Whether the input iterables are sorted in descending
        order.
    :return: Generator yielding all unique items in sorted order.
    """
    merged = heapq.merge(*iterables, key=key, reverse=reverse)
    return (list(group)[0] for _, group in itertools.groupby(merged, key=key))


def intersect_generators(*iterables, allow_duplicates: bool = False):
    """Generator of intersect iterables.

    Yield items only if they are yielded by all iterables. zip_longest
    is used to retrieve items from all iterables in parallel, so that
    items can be yielded before iterables are exhausted.

    Generator is stopped when all iterables are exhausted. Quitting
    before all iterables are finished is attempted if there is no more
    chance of finding an item in all of them.

    Sample:

    >>> iterables = 'mississippi', 'missouri'
    >>> list(intersect_generators(*iterables))
    ['m', 'i', 's']
    >>> list(intersect_generators(*iterables, allow_duplicates=True))
    ['m', 'i', 's', 's', 'i']


    .. versionadded:: 3.0

    .. versionchanged:: 5.0
       Avoid duplicates (:phab:`T263947`).

    .. versionchanged:: 6.4
       ``genlist`` was renamed to ``iterables``; consecutive iterables
       are to be used as iterables parameters or '*' to unpack a list

    .. versionchanged:: 7.0
       Reimplemented without threads which is up to 10'000 times faster

    .. versionchanged:: 9.0
       Iterable elements may consist of lists or tuples
       ``allow_duplicates`` is a keyword-only argument

    :param iterables: page generators
    :param allow_duplicates: optional keyword argument to allow duplicates
        if present in all generators
    """
    if not iterables:
        return

    if len(iterables) == 1:
        yield from iterables[0]
        return

    # If any iterable is empty, no pages are going to be returned
    for source in iterables:
        if not source:
            debug(f'At least one iterable ({source!r}) is empty and execution'
                  ' was skipped immediately.')
            return

    # Item is cached to check that it is found n_gen times
    # before being yielded.
    cache = collections.defaultdict(collections.Counter)
    n_gen = len(iterables)

    ones = collections.Counter(range(n_gen))
    active_iterables = set(range(n_gen))
    seen = set()

    # Get items from iterables in a round-robin way.
    sentinel = object()
    for items in itertools.zip_longest(*iterables, fillvalue=sentinel):
        for index, item in enumerate(items):

            if item is sentinel:
                active_iterables.discard(index)
                continue

            if not allow_duplicates and hash(item) in seen:
                continue

            # Each cache entry is a Counter of iterables' index
            cache[item][index] += 1

            if len(cache[item]) == n_gen:
                yield item

                # Remove item from cache if possible or decrease Counter entry
                if not allow_duplicates:
                    del cache[item]
                    seen.add(hash(item))
                elif cache[item] == ones:
                    del cache[item]
                else:
                    cache[item] -= ones

        # We can quit if an iterable is exceeded and cached iterables is
        # a subset of active iterables.
        if len(active_iterables) < n_gen:
            cached_iterables = set(
                itertools.chain.from_iterable(v.keys()
                                              for v in cache.values()))
            if cached_iterables <= active_iterables:
                return


def roundrobin_generators(*iterables) -> Generator[Any, None, None]:
    """Yield simultaneous from each iterable.

    Sample:

    >>> tuple(roundrobin_generators('ABC', range(5)))
    ('A', 0, 'B', 1, 'C', 2, 3, 4)

    .. versionadded:: 3.0
    .. versionchanged:: 6.4
       A sentinel variable is used to determine the end of an iterable
       instead of None.

    :param iterables: any iterable to combine in roundrobin way
    :type iterables: iterable
    :return: the combined generator of iterables
    :rtype: generator
    """
    sentinel = object()
    return (item
            for item in itertools.chain.from_iterable(
                itertools.zip_longest(*iterables, fillvalue=sentinel))
            if item is not sentinel)


def filter_unique(iterable, container=None, key=None, add=None):
    """Yield unique items from an iterable, omitting duplicates.

    By default, to provide uniqueness, it puts the generated items into a
    set created as a local variable. It only yields items which are not
    already present in the local set.

    For large collections, this is not memory efficient, as a strong reference
    to every item is kept in a local set which cannot be cleared.

    Also, the local set can't be re-used when chaining unique operations on
    multiple generators.

    To avoid these issues, it is advisable for the caller to provide their own
    container and set the key parameter to be the function
    :py:obj:`hash`, or use a :py:obj:`weakref` as the key.

    The container can be any object that supports __contains__.
    If the container is a set or dict, the method add or __setitem__ will be
    used automatically. Any other method may be provided explicitly using the
    add parameter.

    Beware that key=id is only useful for cases where id() is not unique.

    .. warning:: This is not thread safe.

    .. versionadded:: 3.0

    :param iterable: the source iterable
    :type iterable: collections.abc.Iterable
    :param container: storage of seen items
    :type container: type
    :param key: function to convert the item to a key
    :type key: callable
    :param add: function to add an item to the container
    :type add: callable
    """
    if container is None:
        container = set()

    if not add:
        if hasattr(container, 'add'):
            def container_add(x) -> None:
                container.add(key(x) if key else x)

            add = container_add
        else:
            def container_setitem(x) -> None:
                container.__setitem__(key(x) if key else x,
                                      True)

            add = container_setitem

    for item in iterable:
        try:
            if (key(item) if key else item) not in container:
                add(item)
                yield item
        except StopIteration:
            return



"""Objects representing API interface to Wikibase site."""
#
# (C) Pywikibot team, 2012-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import datetime
import json
import uuid
from contextlib import suppress
from typing import Any
from warnings import warn

import pywikibot
from pywikibot.backports import Generator, Iterable, batched
from pywikibot.data import api
from pywikibot.exceptions import (
    APIError,
    EntityTypeUnknownError,
    IsRedirectPageError,
    NoPageError,
    NoWikibaseEntityError,
)
from pywikibot.site._apisite import APISite
from pywikibot.site._decorators import need_extension, need_right
from pywikibot.tools import deprecated, merge_unique_dicts


__all__ = ('DataSite', )


class DataSite(APISite):

    """Wikibase data capable site."""

    def __init__(self, *args, **kwargs) -> None:
        """Initializer."""
        super().__init__(*args, **kwargs)
        self._item_namespace = None
        self._property_namespace = None
        self._type_to_class = {
            'item': pywikibot.ItemPage,
            'property': pywikibot.PropertyPage,
            'mediainfo': pywikibot.MediaInfo,
            'lexeme': pywikibot.LexemePage,
            'form': pywikibot.LexemeForm,
            'sense': pywikibot.LexemeSense,
        }

    def get_repo_for_entity_type(self, entity_type: str) -> DataSite:
        """Get the data repository for the entity type.

        When no foreign repository is defined for the entity type,
        the method returns this repository itself even if it does not
        support that entity type either.

        .. seealso:: https://www.mediawiki.org/wiki/Wikibase/Federation
        .. versionadded:: 8.0

        :raises ValueError: when invalid entity type was provided
        """
        if entity_type not in self._type_to_class:
            raise ValueError(f'Invalid entity type "{entity_type}"')
        entity_sources = self.entity_sources()
        if entity_type in entity_sources:
            return pywikibot.Site(
                *entity_sources[entity_type],
                interface='DataSite',
                user=self.username())
        return self

    def _cache_entity_namespaces(self) -> None:
        """Find namespaces for each known wikibase entity type."""
        self._entity_namespaces = {}
        for entity_type in self._type_to_class:
            for namespace in self.namespaces.values():
                if not hasattr(namespace, 'defaultcontentmodel'):
                    continue

                content_model = namespace.defaultcontentmodel
                if content_model == ('wikibase-' + entity_type):
                    self._entity_namespaces[entity_type] = namespace
                    break

    def get_namespace_for_entity_type(self, entity_type):
        """Return namespace for given entity type.

        :return: corresponding namespace
        :rtype: Namespace
        """
        if not hasattr(self, '_entity_namespaces'):
            self._cache_entity_namespaces()
        if entity_type in self._entity_namespaces:
            return self._entity_namespaces[entity_type]
        raise EntityTypeUnknownError(
            f'{self!r} does not support entity type "{entity_type}" '
            " or it doesn't have its own namespace"
        )

    @property
    def item_namespace(self):
        """Return namespace for items.

        :return: item namespace
        :rtype: Namespace
        """
        if self._item_namespace is None:
            self._item_namespace = self.get_namespace_for_entity_type('item')
        return self._item_namespace

    @property
    def property_namespace(self):
        """Return namespace for properties.

        :return: property namespace
        :rtype: Namespace
        """
        if self._property_namespace is None:
            self._property_namespace = self.get_namespace_for_entity_type(
                'property')
        return self._property_namespace

    def get_entity_for_entity_id(self, entity_id):
        """Return a new instance for given entity id.

        :raises pywikibot.exceptions.NoWikibaseEntityError: there is no
            entity with the id
        :return: a WikibaseEntity subclass
        :rtype: WikibaseEntity
        """
        for cls in self._type_to_class.values():
            if cls.is_valid_id(entity_id):
                return cls(self, entity_id)

        entity = pywikibot.page.WikibaseEntity(self, entity_id)
        raise NoWikibaseEntityError(entity)

    @property
    def sparql_endpoint(self) -> str | None:
        """Return the sparql endpoint url, if any has been set.

        :return: sparql endpoint url
        """
        return self.siteinfo.get('wikibase-sparql')

    @property
    def concept_base_uri(self) -> str:
        """Return the base uri for concepts/entities.

        :return: concept base uri
        """
        return self.siteinfo['wikibase-conceptbaseuri']

    def geo_shape_repository(self) -> DataSite | None:
        """Return Site object for the geo-shapes repository e.g. commons."""
        url = self.siteinfo.get('wikibase-geoshapestoragebaseurl')
        if url:
            return pywikibot.Site(url=url, user=self.username())

        return None

    def tabular_data_repository(self) -> DataSite | None:
        """Return Site object for the tabular-data repository e.g. commons."""
        url = self.siteinfo.get('wikibase-tabulardatastoragebaseurl')
        if url:
            return pywikibot.Site(url=url, user=self.username())

        return None

    def loadcontent(self, identification, *props):
        """Fetch the current content of a Wikibase item.

        This is called loadcontent since wbgetentities does not support
        fetching old revisions. Eventually this will get replaced by an
        actual loadrevisions.

        :param identification: Parameters used to identify the page(s)
        :type identification: dict
        :param props: the optional properties to fetch.
        """
        params = merge_unique_dicts(identification, action='wbgetentities',
                                    # TODO: When props is empty it results in
                                    # an empty string ('&props=') but it should
                                    # result in a missing entry.
                                    props=props or False)
        req = self.simple_request(**params)
        data = req.submit()
        if 'success' not in data:
            raise APIError(data['errors'], '')
        return data['entities']

    def preload_entities(
        self,
        pagelist: Iterable[pywikibot.page.WikibaseEntity
                           | pywikibot.page.Page],
        groupsize: int = 50
    ) -> Generator[pywikibot.page.WikibaseEntity, None, None]:
        """Yield subclasses of WikibaseEntity's with content prefilled.

        .. note:: Pages will be iterated in a different order than in
           the underlying pagelist.

        :param pagelist: an iterable that yields either WikibaseEntity
            objects, or Page objects linked to an ItemPage.
        :param groupsize: how many pages to query at a time
        """
        if not hasattr(self, '_entity_namespaces'):
            self._cache_entity_namespaces()
        for batch in batched(pagelist, groupsize):
            req: dict[str, list[str]] = {'ids': [], 'titles': [], 'sites': []}
            for p in batch:
                if isinstance(p, pywikibot.page.WikibaseEntity):
                    ident = p._defined_by()
                    for key in ident:
                        req[key].append(ident[key])
                elif (p.site == self
                      and p.namespace() in self._entity_namespaces.values()):
                    req['ids'].append(p.title(with_ns=False))
                else:
                    assert p.site.has_data_repository, \
                        'Site must have a data repository'
                    req['sites'].append(p.site.dbName())
                    req['titles'].append(p._link._text)

            req = self.simple_request(action='wbgetentities', **req)
            data = req.submit()
            for entity in data['entities']:
                if 'missing' in data['entities'][entity]:
                    continue
                cls = self._type_to_class[data['entities'][entity]['type']]
                page = cls(self, entity)
                # No api call is made because item._content is given
                page._content = data['entities'][entity]
                with suppress(IsRedirectPageError):
                    page.get()  # cannot provide get_redirect=True (T145971)
                yield page

    def get_property_type(self, prop: pywikibot.page.Property) -> str:
        """Obtain the type of a property.

        This is used specifically because we can cache the value for a
        much longer time (near infinite).

        .. versionadded:: 9.5

        :raises NoWikibaseEntityError: *prop* does not exist
        """
        params = {'action': 'wbgetentities', 'ids': prop.getID(),
                  'props': 'datatype'}
        expiry = datetime.timedelta(days=365 * 100)
        # Store it for 100 years
        req = self._request(expiry=expiry, parameters=params)
        data = req.submit()

        # the IDs returned from the API can be upper or lowercase, depending
        # on the version. See bug T55894 for more information.
        try:
            entity = data['entities'][prop.getID()]
        except KeyError:
            entity = data['entities'][prop.getID().lower()]

        if 'missing' in entity:
            raise NoWikibaseEntityError(
                prop if isinstance(prop, pywikibot.page.WikibaseEntity)
                else pywikibot.page.WikibaseEntity(self, prop.getID())
            )

        return entity['datatype']

    @deprecated('get_property_type', since='9.5.0')
    def getPropertyType(self, prop):
        """Obtain the type of a property.

        .. deprecated:: 9.5
           Use :meth:`get_property_type` instead.
        """
        try:
            return self.get_property_type(prop)
        except NoWikibaseEntityError as exc:
            raise KeyError(f'{exc.entity.id} does not exist') from None

    @need_right('edit')
    def editEntity(self,
                   entity: pywikibot.page.WikibaseEntity | dict,
                   data: dict,
                   bot: bool = True,
                   **kwargs) -> dict:
        """Edit entity.

        .. note:: This method is unable to create entities other than
           ``item`` if dict with API parameters was passed to *entity*
           parameter.

        .. versionchanged:: 9.4
           *tags* keyword argument was added

        :param entity: Page to edit, or dict with API parameters
            to use for entity identification.
        :param data: data updates
        :param bot: Whether to mark the edit as a bot edit.

        :keyword int baserevid: The numeric identifier for the revision
            to base the modification on. This is used for detecting
            conflicts during save.
        :keyword bool clear: If set, the complete entity is emptied
            before proceeding. The entity will not be saved before it is
            filled with the *data*, possibly with parts excluded.
        :keyword str summary: Summary for the edit. Will be prepended by
            an automatically generated comment. The length limit of the
            autocomment together with the summary is 260 characters. Be
            aware that everything above that limit will be cut off.
        :keyword Iterable[str] | str tags: Change tags to apply to the
            revision.
        :return: New entity data
        """
        # this changes the reference to a new object
        data = dict(data)
        if isinstance(entity, pywikibot.page.WikibaseEntity):
            params = entity._defined_by(singular=True)
            if 'id' in params and params['id'] == '-1':
                del params['id']
            if not params:
                params['new'] = entity.entity_type
                data_for_new_entity = entity.get_data_for_new_entity()
                data.update(data_for_new_entity)
        else:
            if 'id' in entity and entity['id'] == '-1':
                del entity['id']
            params = dict(entity)
            if not params:  # If no identification was provided
                params['new'] = 'item'

        params.update(action='wbeditentity', bot=bot)

        if kwargs.get('baserevid'):
            params['baserevid'] = kwargs['baserevid']

        params['token'] = self.tokens['csrf']

        for arg, param in kwargs.items():
            if arg in ['clear', 'summary', 'tags']:
                params[arg] = param
            elif arg != 'baserevid':
                warn(f'Unknown wbeditentity parameter {arg} ignored',
                     UserWarning, 2)

        params['data'] = json.dumps(data)
        req = self.simple_request(**params)
        return req.submit()

    @need_right('edit')
    def addClaim(self,
                 entity: pywikibot.page.WikibaseEntity,
                 claim: pywikibot.page.Claim,
                 bot: bool = True,
                 summary: str | None = None,
                 tags: str | None = None) -> None:
        """Add a claim.

        .. versionchanged:: 9.4
           *tags* parameter was added

        :param entity: Entity to modify
        :param claim: Claim to be added
        :param bot: Whether to mark the edit as a bot edit
        :param summary: Edit summary
        :param tags: Change tags to apply to the revision
        """
        claim.snak = entity.getID() + '$' + str(uuid.uuid4())
        params = {
            'action': 'wbsetclaim',
            'claim': json.dumps(claim.toJSON()),
            'baserevid': entity.latest_revision_id,
            'summary': summary,
            'tags': tags,
            'bot': bot,
            'token': self.tokens['csrf'],
        }
        req = self.simple_request(**params)
        data = req.submit()
        # Update the item
        if claim.getID() in entity.claims:
            entity.claims[claim.getID()].append(claim)
        else:
            entity.claims[claim.getID()] = [claim]
        entity.latest_revision_id = data['pageinfo']['lastrevid']

    @need_right('edit')
    def changeClaimTarget(self,
                          claim: pywikibot.Claim,
                          snaktype: str = 'value',
                          bot: bool = True,
                          summary: str | None = None,
                          tags: str | None = None):
        """Set the claim target to the value of the provided claim target.

        .. versionchanged:: 9.4
           *tags* parameter was added

        :param claim: The source of the claim target value
        :param snaktype: An optional snaktype ('value', 'novalue' or
            'somevalue'). Default: 'value'
        :param bot: Whether to mark the edit as a bot edit
        :param summary: Edit summary
        :param tags: Change tags to apply to the revision
        """
        if claim.isReference or claim.isQualifier:
            raise NotImplementedError

        if not claim.snak:
            # We need to already have the snak value
            raise NoPageError(claim)

        params = {
            'action': 'wbsetclaimvalue',
            'claim': claim.snak,
            'snaktype': snaktype,
            'summary': summary,
            'tags': tags,
            'bot': bot,
            'token': self.tokens['csrf'],
        }

        if snaktype == 'value':
            params['value'] = json.dumps(claim._formatValue())

        params['baserevid'] = claim.on_item.latest_revision_id
        req = self.simple_request(**params)
        return req.submit()

    @need_right('edit')
    def save_claim(self,
                   claim: pywikibot.page.Claim,
                   summary: str | None = None,
                   bot: bool = True,
                   tags: str | None = None):
        """Save the whole claim to the wikibase site.

        .. versionchanged:: 9.4
           *tags* parameter was added

        :param claim: The claim to save
        :param bot: Whether to mark the edit as a bot edit
        :param summary: Edit summary
        :param tags: Change tags to apply to the revision
        :raises NoPageError: missing the the snak value
        :raises NotImplementedError: ``claim.isReference`` or
            ``claim.isQualifier`` is given
        """
        if claim.isReference or claim.isQualifier:
            raise NotImplementedError

        if not claim.snak:
            # We need to already have the snak value
            raise NoPageError(claim)

        params = {
            'action': 'wbsetclaim',
            'claim': json.dumps(claim.toJSON()),
            'baserevid': claim.on_item.latest_revision_id,
            'summary': summary,
            'tags': tags,
            'bot': bot,
            'token': self.tokens['csrf'],
        }

        req = self.simple_request(**params)
        data = req.submit()
        claim.on_item.latest_revision_id = data['pageinfo']['lastrevid']
        return data

    @need_right('edit')
    def editSource(self,
                   claim: pywikibot.Claim,
                   source: pywikibot.Claim,
                   new: bool = False,
                   bot: bool = True,
                   summary: str | None = None,
                   tags: str | None = None):
        """Create/Edit a source.

        .. versionchanged:: 9.4
           *tags* parameter was added
        .. versionchanged:: 10.0
           deprecated *baserevid* parameter was removed

        :param claim: A Claim object to add the source to.
        :param source: A Claim object to be used as a source.
        :param new: Whether to create a new one if the "source" already
            exists.
        :param bot: Whether to mark the edit as a bot edit.
        :param summary: Edit summary.
        :param tags: Change tags to apply to the revision.
        :raises ValueError: The claim cannot have a source.
        """
        if claim.isReference or claim.isQualifier:
            raise ValueError('The claim cannot have a source.')

        params = {
            'action': 'wbsetreference',
            'statement': claim.snak,
            'baserevid': claim.on_item.latest_revision_id,
            'summary': summary,
            'tags': tags,
            'bot': bot,
            'token': self.tokens['csrf'],
        }

        # build up the snak
        sources = source if isinstance(source, list) else [source]

        snak = {}
        for sourceclaim in sources:
            datavalue = sourceclaim._formatDataValue()
            valuesnaks = snak.get(sourceclaim.getID(), [])
            valuesnaks.append({
                'snaktype': 'value',
                'property': sourceclaim.getID(),
                'datavalue': datavalue,
            })

            snak[sourceclaim.getID()] = valuesnaks
            # set the hash if the source should be changed.
            # if present, all claims of one source have the same hash
            if not new and hasattr(sourceclaim, 'hash'):
                params['reference'] = sourceclaim.hash
        params['snaks'] = json.dumps(snak)

        req = self.simple_request(**params)
        return req.submit()

    @need_right('edit')
    def editQualifier(self,
                      claim: pywikibot.Claim,
                      qualifier: pywikibot.Claim,
                      new: bool = False,
                      bot: bool = True,
                      summary: str | None = None,
                      tags: str | None = None):
        """Create/Edit a qualifier.

        .. versionchanged:: 9.4
           *tags* parameter was added
        .. versionchanged:: 10.0
           deprecated *baserevid* parameter was removed

        :param claim: A Claim object to add the qualifier to
        :param qualifier: A Claim object to be used as a qualifier
        :param new: Whether to create a new one if the qualifier
            already exists
        :param bot: Whether to mark the edit as a bot edit
        :param summary: Edit summary
        :param tags: Change tags to apply to the revision
        :raises ValueError: The claim cannot have a qualifier.
        """
        if claim.isReference or claim.isQualifier:
            raise ValueError('The claim cannot have a qualifier.')

        params = {
            'action': 'wbsetqualifier',
            'claim': claim.snak,
            'baserevid': claim.on_item.latest_revision_id,
            'summary': summary,
            'tags': tags,
            'bot': bot,
            'token': self.tokens['csrf'],
        }

        if (not new and hasattr(qualifier, 'hash')
                and qualifier.hash is not None):
            params['snakhash'] = qualifier.hash

        # build up the snak
        if qualifier.getSnakType() == 'value':
            params['value'] = json.dumps(qualifier._formatValue())

        params['snaktype'] = qualifier.getSnakType()
        params['property'] = qualifier.getID()

        req = self.simple_request(**params)
        return req.submit()

    @need_right('edit')
    def removeClaims(self,
                     claims: list[pywikibot.Claim],
                     bot: bool = True,
                     summary: str | None = None,
                     tags: str | None = None):
        """Remove claims.

        .. versionchanged:: 9.4
           *tags* parameter was added
        .. versionchanged:: 10.0
           deprecated *baserevid* parameter was removed

        :param claims: Claims to be removed
        :param bot: Whether to mark the edit as a bot edit
        :param summary: Edit summary
        :param tags: Change tags to apply to the revision
        """
        # Check on_item for all additional claims
        items = {claim.on_item for claim in claims if claim.on_item}
        assert len(items) == 1
        baserevid = items.pop().latest_revision_id

        params = {
            'action': 'wbremoveclaims',
            'claim': '|'.join(claim.snak for claim in claims),
            'baserevid': baserevid,
            'summary': summary,
            'tags': tags,
            'bot': bot,
            'token': self.tokens['csrf'],
        }

        req = self.simple_request(**params)
        return req.submit()

    @need_right('edit')
    def removeSources(self,
                      claim: pywikibot.Claim,
                      sources: list[pywikibot.Claim],
                      bot: bool = True,
                      summary: str | None = None,
                      tags: str | None = None):
        """Remove sources.

        .. versionchanged:: 9.4
           *tags* parameter was added
        .. versionchanged:: 10.0
           deprecated `baserevid` parameter was removed

        :param claim: A Claim object to remove the sources from
        :param sources: A list of Claim objects that are sources
        :param bot: Whether to mark the edit as a bot edit
        :param summary: Edit summary
        :param tags: Change tags to apply to the revision
        """
        params = {
            'action': 'wbremovereferences',
            'statement': claim.snak,
            'references': '|'.join(source.hash for source in sources),
            'baserevid': claim.on_item.latest_revision_id,
            'summary': summary,
            'tags': tags,
            'bot': bot,
            'token': self.tokens['csrf'],
        }

        req = self.simple_request(**params)
        return req.submit()

    @need_right('edit')
    def remove_qualifiers(self,
                          claim: pywikibot.Claim,
                          qualifiers: list[pywikibot.Claim],
                          bot: bool = True,
                          summary: str | None = None,
                          tags: str | None = None):
        """Remove qualifiers.

        .. versionchanged:: 9.4
           *tags* parameter was added
        .. versionchanged:: 10.0
           deprecated `baserevid` parameter was removed

        :param claim: A Claim object to remove the qualifier from
        :param qualifiers: Claim objects currently used as a qualifiers
        :param bot: Whether to mark the edit as a bot edit
        :param summary: Edit summary
        :param tags: Change tags to apply to the revision
        """
        params = {
            'action': 'wbremovequalifiers',
            'claim': claim.snak,
            'qualifiers': [qualifier.hash for qualifier in qualifiers],
            'baserevid': claim.on_item.latest_revision_id,
            'summary': summary,
            'tags': tags,
            'bot': bot,
            'token': self.tokens['csrf'],
        }

        req = self.simple_request(**params)
        return req.submit()

    @need_right('edit')
    def linkTitles(self,
                   page1: pywikibot.Page,
                   page2: pywikibot.Page,
                   bot: bool = True) -> dict:
        """Link two pages together.

        .. versionchanged:: 9.4
           *tags* parameter was added

        :param page1: First page to link
        :param page2: Second page to link
        :param bot: Whether to mark the edit as a bot edit
        :return: dict API output
        """
        params = {
            'action': 'wblinktitles',
            'tosite': page1.site.dbName(),
            'totitle': page1.title(),
            'fromsite': page2.site.dbName(),
            'fromtitle': page2.title(),
            'bot': bot,
            'token': self.tokens['csrf']
        }

        req = self.simple_request(**params)
        return req.submit()

    @need_right('item-merge')
    def mergeItems(self,
                   from_item: pywikibot.ItemPage,
                   to_item: pywikibot.ItemPage,
                   ignore_conflicts: list[str] | None = None,
                   summary: str | None = None,
                   bot: bool = True,
                   tags: str | None = None) -> dict:
        """Merge two items together.

        .. versionchanged:: 9.4
           *tags* parameter was added

        :param from_item: Item to merge from
        :param to_item: Item to merge into
        :param ignore_conflicts: Which type of conflicts
            ('description', 'sitelink', and 'statement')
            should be ignored
        :param summary: Edit summary
        :param bot: Whether to mark the edit as a bot edit
        :param tags: Change tags to apply to the revision
        :return: dict API output
        """
        params = {
            'action': 'wbmergeitems',
            'fromid': from_item.getID(),
            'toid': to_item.getID(),
            'ignoreconflicts': ignore_conflicts,
            'summary': summary,
            'tags': tags,
            'bot': bot,
            'token': self.tokens['csrf'],
        }

        req = self.simple_request(**params)
        return req.submit()

    @need_right('item-merge')
    @need_extension('WikibaseLexeme')
    def mergeLexemes(self, from_lexeme, to_lexeme, summary=None, *,
                     bot: bool = True) -> dict:
        """Merge two lexemes together.

        :param from_lexeme: Lexeme to merge from
        :type from_lexeme: pywikibot.LexemePage
        :param to_lexeme: Lexeme to merge into
        :type to_lexeme: pywikibot.LexemePage
        :param summary: Edit summary
        :type summary: str
        :keyword bot: Whether to mark the edit as a bot edit
        :return: dict API output
        """
        params = {
            'action': 'wblmergelexemes',
            'source': from_lexeme.getID(),
            'target': to_lexeme.getID(),
            'token': self.tokens['csrf'],
            'summary': summary,
            'bot': bot,
        }

        req = self.simple_request(**params)
        return req.submit()

    @need_right('item-redirect')
    def set_redirect_target(self, from_item, to_item, bot: bool = True):
        """Make a redirect to another item.

        :param to_item: title of target item.
        :type to_item: pywikibot.ItemPage
        :param from_item: Title of the item to be redirected.
        :type from_item: pywikibot.ItemPage
        :param bot: Whether to mark the edit as a bot edit
        """
        params = {
            'action': 'wbcreateredirect',
            'from': from_item.getID(),
            'to': to_item.getID(),
            'token': self.tokens['csrf'],
            'bot': bot,
        }
        req = self.simple_request(**params)
        return req.submit()

    def search_entities(self, search: str, language: str,
                        total: int | None = None, **kwargs):
        """Search for pages or properties that contain the given text.

        :param search: Text to find.
        :param language: Language to search in.
        :param total: Maximum number of pages to retrieve in total, or
            None in case of no limit.
        :return: 'search' list from API output.
        :rtype: Generator
        """
        lang_codes = self._paraminfo.parameter('wbsearchentities',
                                               'language')['type']
        if language not in lang_codes:
            raise ValueError('Data site used does not support provided '
                             'language.')

        if 'site' in kwargs:
            if kwargs['site'].sitename != self.sitename:
                raise ValueError('The site given in the kwargs is different.')

            warn('search_entities should not get a site via kwargs.',
                 UserWarning, 2)
            del kwargs['site']

        parameters = dict(search=search, language=language, **kwargs)
        return self._generator(api.APIGenerator,
                               type_arg='wbsearchentities',
                               data_name='search',
                               total=total, parameters=parameters)

    def parsevalue(self, datatype: str, values: list[str],
                   options: dict[str, Any] | None = None,
                   language: str | None = None,
                   validate: bool = False) -> list[Any]:
        """Send data values to the wikibase parser for interpretation.

        .. versionadded:: 7.5
        .. seealso:: `wbparsevalue API
           <https://www.wikidata.org/w/api.php?action=help&modules=wbparsevalue>`_

        :param datatype: datatype of the values being parsed. Refer the
            API for a valid datatype.
        :param values: list of values to be parsed
        :param options: any additional options for wikibase parser
            (for time, 'precision' should be specified)
        :param language: code of the language to parse the value in
        :param validate: whether parser should provide data validation as well
            as parsing
        :return: list of parsed values
        :raises ValueError: parsing failed due to some invalid input values
        """
        params = {
            'action': 'wbparsevalue',
            'datatype': datatype,
            'values': values,
            'options': json.dumps(options or {}),
            'validate': validate,
            'uselang': language or 'en',
        }
        req = self.simple_request(**params)
        try:
            data = req.submit()
        except APIError as e:
            if e.code.startswith('wikibase-parse-error'):
                for err in e.other['results']:
                    if 'error' in err:
                        pywikibot.error('{error-info} for value {raw!r}, '
                                        '{expected-format!r} format expected'
                                        .format_map(err))
                raise ValueError(e) from None
            raise

        if 'results' not in data:
            raise RuntimeError(
                f"Unexpected missing 'results' in query data\n{data}")

        results = []
        for result_hash in data['results']:
            if 'value' not in result_hash:
                # There should be an APIError occurred already
                raise RuntimeError("Unexpected missing 'value' in query data:"
                                   f'\n{result_hash}')
            results.append(result_hash['value'])
        return results

    @need_right('edit')
    def _wbset_action(self, itemdef, action: str, action_data,
                      **kwargs) -> dict:
        """Execute wbset{action} on a Wikibase entity.

        Supported actions are:
            wbsetaliases, wbsetdescription, wbsetlabel and wbsetsitelink

        wbsetaliases:
            dict shall have the following structure:

            .. code-block:: python

               {
                   'language': value (str),
                   'add': list of language codes (str),
                   'remove': list of language codes (str),
                   'set' list of language codes (str),
               }

            'add' and 'remove' are alternative to 'set'

        wbsetdescription and wbsetlabel:
            dict shall have keys 'language', 'value'

        wbsetsitelink:
            dict shall have keys 'linksite', 'linktitle' and
            optionally 'badges'

        :param itemdef: Entity to modify or create
        :type itemdef: str, WikibaseEntity or Page connected to such item
        :param action: wbset{action} to perform:
            'wbsetaliases', 'wbsetdescription', 'wbsetlabel', 'wbsetsitelink'
        :param action_data: data to be used in API request, see API help
        :type action_data: SiteLink or dict
        :keyword bot: Whether to mark the edit as a bot edit, default is True
        :type bot: bool
        :keyword tags: Change tags to apply with the edit
        :type tags: list of str
        :return: query result
        :raises: AssertionError, TypeError
        """
        def format_sitelink(sitelink):
            """Convert SiteLink to a dict accepted by wbsetsitelink API."""
            if isinstance(sitelink, pywikibot.page.SiteLink):
                _dict = {
                    'linksite': sitelink._sitekey,
                    'linktitle': sitelink._rawtitle,
                    'badges': '|'.join([b.title() for b in sitelink.badges]),
                }
            else:
                _dict = sitelink

            return _dict

        def prepare_data(action, data):
            """Prepare data as expected by API."""
            if action == 'wbsetaliases':
                res = data
                keys = set(res)
                assert keys < {'language', 'add', 'remove', 'set'}
                assert 'language' in keys
                assert ({'add', 'remove', 'set'} & keys)
                assert ({'add', 'set'} >= keys)
                assert ({'remove', 'set'} >= keys)
            elif action in ('wbsetlabel', 'wbsetdescription'):
                res = data
                keys = set(res)
                assert keys == {'language', 'value'}
            elif action == 'wbsetsitelink':
                res = format_sitelink(data)
                keys = set(res)
                assert keys >= {'linksite'}
                assert keys <= {'linksite', 'linktitle', 'badges'}
            else:
                raise ValueError('Something has gone wrong ...')

            return res

        # Supported actions
        assert action in ('wbsetaliases', 'wbsetdescription',
                          'wbsetlabel', 'wbsetsitelink'), \
            f'action {action} not supported.'

        # prefer ID over (site, title)
        if isinstance(itemdef, str):
            itemdef = self.get_entity_for_entity_id(itemdef)
        elif isinstance(itemdef, pywikibot.Page):
            itemdef = pywikibot.ItemPage.fromPage(itemdef, lazy_load=True)
        elif not isinstance(itemdef, pywikibot.page.WikibaseEntity):
            raise TypeError('itemdef shall be str, WikibaseEntity or Page')

        params = itemdef._defined_by(singular=True)
        # TODO: support 'new'
        baserevid = kwargs.pop(
            'baserevid',
            itemdef.latest_revision_id if 'id' in params else 0
        )
        params.update(
            {'baserevid': baserevid,
             'action': action,
             'token': self.tokens['csrf'],
             'bot': kwargs.pop('bot', True),
             })
        params.update(prepare_data(action, action_data))

        for arg, param in kwargs.items():
            if arg in ['summary', 'tags']:
                params[arg] = param
            else:
                warn(f'Unknown parameter {arg} for action {action}, ignored',
                     UserWarning, 2)

        req = self.simple_request(**params)
        return req.submit()

    def wbsetaliases(self, itemdef, aliases, **kwargs):
        """Set aliases for a single Wikibase entity.

        See self._wbset_action() for parameters
        """
        return self._wbset_action(itemdef, 'wbsetaliases', aliases, **kwargs)

    def wbsetdescription(self, itemdef, description, **kwargs):
        """Set description for a single Wikibase entity.

        See self._wbset_action()
        """
        return self._wbset_action(itemdef, 'wbsetdescription', description,
                                  **kwargs)

    def wbsetlabel(self, itemdef, label, **kwargs):
        """Set label for a single Wikibase entity.

        See self._wbset_action() for parameters
        """
        return self._wbset_action(itemdef, 'wbsetlabel', label, **kwargs)

    def wbsetsitelink(self, itemdef, sitelink, **kwargs):
        """Set, remove or modify a sitelink on a Wikibase item.

        See self._wbset_action() for parameters
        """
        return self._wbset_action(itemdef, 'wbsetsitelink', sitelink, **kwargs)

    @need_right('edit')
    @need_extension('WikibaseLexeme')
    def add_form(self, lexeme, form, *, bot: bool = True,
                 baserevid=None) -> dict:
        """Add a form.

        :param lexeme: Lexeme to modify
        :type lexeme: pywikibot.LexemePage
        :param form: Form to be added
        :type form: pywikibot.LexemeForm
        :keyword bot: Whether to mark the edit as a bot edit
        :keyword baserevid: Base revision id override, used to detect
            conflicts.
        :type baserevid: long
        """
        params = {
            'action': 'wbladdform',
            'lexemeId': lexeme.getID(),
            'data': json.dumps(form.toJSON()),
            'bot': bot,
            'token': self.tokens['csrf'],
        }
        if baserevid:
            params['baserevid'] = baserevid
        req = self.simple_request(**params)
        return req.submit()

    @need_right('edit')
    @need_extension('WikibaseLexeme')
    def remove_form(self, form, *, bot: bool = True, baserevid=None) -> dict:
        """Remove a form.

        :param form: Form to be removed
        :type form: pywikibot.LexemeForm
        :keyword bot: Whether to mark the edit as a bot edit
        :keyword baserevid: Base revision id override, used to detect
            conflicts.
        :type baserevid: long
        """
        params = {
            'action': 'wblremoveform',
            'id': form.getID(),
            'bot': bot,
            'token': self.tokens['csrf'],
        }
        if baserevid:
            params['baserevid'] = baserevid
        req = self.simple_request(**params)
        return req.submit()

    @need_right('edit')
    @need_extension('WikibaseLexeme')
    def edit_form_elements(self, form, data, *, bot: bool = True,
                           baserevid=None) -> dict:
        """Edit lexeme form elements.

        :param form: Form
        :type form: pywikibot.LexemeForm
        :param data: data updates
        :type data: dict
        :keyword bot: Whether to mark the edit as a bot edit
        :keyword baserevid: Base revision id override, used to detect
            conflicts.
        :type baserevid: long
        :return: New form data
        """
        params = {
            'action': 'wbleditformelements',
            'formId': form.getID(),
            'data': json.dumps(data),
            'bot': bot,
            'token': self.tokens['csrf'],
        }
        if baserevid:
            params['baserevid'] = baserevid
        req = self.simple_request(**params)
        return req.submit()



"""Logging tools."""
#
# (C) Pywikibot team, 2009-2024
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import logging

from pywikibot.userinterfaces.terminal_interface_base import colorTagR


class LoggingFormatter(logging.Formatter):

    """Format LogRecords for output to file."""

    def format(self, record):
        """Strip trailing newlines before outputting text to file."""
        # Warnings captured from the warnings system are not processed by
        # logoutput(), so the 'context' variables are missing.
        if record.name == 'py.warnings' \
           and 'caller_file' not in record.__dict__:
            assert len(record.args) == 1, \
                'Arguments for record is not correctly set'
            msg = record.args[0]

            record.__dict__['caller_file'] = record.filename
            record.__dict__['caller_name'] = record.module
            record.__dict__['caller_line'] = record.lineno

            record.args = (msg,)

        # remove color tags
        if record.msg and isinstance(record.msg, str):
            record.msg = colorTagR.sub('', record.msg)

        return super().format(record).rstrip()



"""Objects representing site info data contents."""
#
# (C) Pywikibot team, 2008-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import copy
import datetime
import re
from collections.abc import Container
from contextlib import suppress
from typing import TYPE_CHECKING, Any, Literal, cast

import pywikibot
from pywikibot.backports import Dict, List
from pywikibot.exceptions import APIError
from pywikibot.tools.collections import EMPTY_DEFAULT


if TYPE_CHECKING:
    from pywikibot.site import APISite


class Siteinfo(Container):

    """A dictionary-like container for siteinfo.

    This class queries the server to get the requested siteinfo
    property. Results can be cached in the instance to avoid repeated
    queries.

    All values of the 'general' property  are directly available.

    .. versionchanged:: 10.5
       formatversion 2 is used for API calls.

    .. admonition:: Compatibility note
       :class: note

       For formatversion 2, some siteinfo data structures differ from
       version 1. Fallback '*' keys are added in the data structure for
       'namespaces', 'languages', 'namespacealiases' and 'skins'
       properties for backwards compatibility. These fallbacks may be
       removed in future versions of Pywikibot.

       The 'thumblimits', 'imagelimits' and 'magiclinks' entries of the
       'general' property are normalized to lists for easier use and to
       match the format used in formatversion 1. For example:

       :code:`'thumblimits': [120, 150, 180, 200, 220, 250, 300, 400]`

    .. deprecated:: 10.5
       Accessing the fallback '*' keys in 'languages', 'namespaces',
       'namespacealiases', and 'skins' properties are deprecated and
       will be removed in a future release of Pywikibot.

    .. seealso:: :api:`siteinfo`
    """

    WARNING_REGEX = re.compile(r'Unrecognized values? for parameter '
                               r'["\']siprop["\']: (.+?)\.?')

    def __init__(self, site: APISite) -> None:
        """Initialize Siteinfo for a given site with an empty cache."""
        self._site = site
        self._cache: dict[str,
                          tuple[Any, datetime.datetime | Literal[False]]] = {}

    def clear(self) -> None:
        """Clear all cached siteinfo properties.

        .. versionadded:: 7.1
        """
        self._cache.clear()

    @staticmethod
    def _post_process(prop: str,
                      data: dict[str, Any] | list[dict[str, Any]]) -> None:
        """Convert empty-string boolean properties to actual booleans.

        Modifies *data* in place.

        .. versionchanged:: 10.5
           Modify *data* for formatversion 1 compatibility and easier
           to use lists.

        :param prop: The siteinfo property name (e.g., 'general',
            'namespaces', 'magicwords')
        :param data: The raw data returned from the server

        :meta public:
        """
        # Be careful with version tests inside this here as it might need to
        # query this method to actually get the version number

        if prop == 'general':
            data = cast(Dict[str, Any], data)
            for key in 'thumblimits', 'imagelimits':
                data[key] = list(data[key].values())
            data['magiclinks'] = [k for k, v in data['magiclinks'].items()
                                  if v]
        elif prop == 'namespaces':
            data = cast(Dict[str, Any], data)
            for ns_info in data.values():
                ns_info['*'] = ns_info['name']
        elif prop in ('languages', 'namespacealiases'):
            data = cast(List[Dict[str, Any]], data)
            for ns_info in data:
                key = 'name' if 'name' in ns_info else 'alias'
                ns_info['*'] = ns_info[key]
        elif prop == 'skins':
            data = cast(List[Dict[str, Any]], data)
            for ns_info in data:
                ns_info['*'] = ns_info['name']
                for key in 'default', 'unusable':
                    ns_info.setdefault(key, False)

    def _get_siteinfo(self, prop, expiry) -> dict:
        """Retrieve one or more siteinfo properties from the server.

        .. seealso:: :api:Siteinfo

        :param prop: The property names of the siteinfo.
        :type prop: str or iterable
        :param expiry: The expiry date of the cached request.
        :type expiry: int (days), :py:obj:`datetime.timedelta`, False (config)
        :return: A dictionary with the properties of the site. Each entry in
            the dictionary is a tuple of the value and a boolean to save if it
            is the default value.
        """
        invalid_properties: list[str] = []

        def warn_handler(mod, message) -> bool:
            """Return True if the warning is handled."""
            matched = Siteinfo.WARNING_REGEX.fullmatch(message)
            if mod == 'siteinfo' and matched:
                invalid_properties.extend(
                    prop.strip() for prop in matched[1].split(','))
                return True
            return False

        # Convert to list for consistent iteration
        props = [prop] if isinstance(prop, str) else list(prop)
        if not props:
            raise ValueError('At least one property name must be provided.')

        request = self._site._request(
            expiry=pywikibot.config.API_config_expiry
            if expiry is False else expiry,
            parameters={
                'action': 'query',
                'meta': 'siteinfo',
                'siprop': props,
                'formatversion': 2,
            }
        )

        # warnings are handled later
        request._warning_handler = warn_handler

        try:
            data = request.submit()
        except APIError as e:
            if e.code == 'siunknown_siprop':
                if len(props) == 1:
                    pywikibot.log(f"Unable to get siprop '{props[0]}'")
                    return {props[0]: (EMPTY_DEFAULT, False)}

                pywikibot.log('Unable to get siteinfo, because at least '
                              "one property is unknown: '{}'"
                              .format("', '".join(props)))
                results = {}
                for p in props:
                    results.update(self._get_siteinfo(p, expiry))
                return results
            raise

        result: dict[str, tuple[Any, datetime.datetime | Literal[False]]] = {}
        if invalid_properties:
            for invalid_prop in invalid_properties:
                result[invalid_prop] = (EMPTY_DEFAULT, False)
            pywikibot.log("Unable to get siprop(s) '{}'"
                          .format("', '".join(invalid_properties)))

        # Process valid properties
        if 'query' in data:
            # If the request is a CachedRequest, use the _cachetime attr.
            cache_time = getattr(
                request, '_cachetime', None) or pywikibot.Timestamp.nowutc()
            for prop in props:
                if prop in data['query']:
                    self._post_process(prop, data['query'][prop])
                    result[prop] = (data['query'][prop], cache_time)
        return result

    @staticmethod
    def _is_expired(cache_date: datetime.datetime | Literal[False] | None,
                    expire: datetime.timedelta | Literal[False]) -> bool:
        """Return true if the cache date is expired.

        :param cache_date: The timestamp when the value was cached, or
            False if default, None if never.
        :param expire: Expiry period as timedelta, or False to never
            expire.
        :return: True if expired, False otherwise.
        """
        if isinstance(expire, bool):
            return expire

        if not cache_date:  # default values are always expired
            return True

        # cached date + expiry are in the past if it's expired
        return cache_date + expire < pywikibot.Timestamp.nowutc()

    def _get_general(self, key: str, expiry):
        """Return a siteinfo property which is loaded by default.

        The property 'general' will be queried if it wasn't yet or it's forced.
        Additionally all uncached default properties are queried. This way
        multiple default properties are queried with one request. It'll cache
        always all results.

        :param key: The key to search for.
        :param expiry: If the cache is older than the expiry it ignores the
            cache and queries the server to get the newest value.
        :type expiry: int (days), :py:obj:`datetime.timedelta`, False (never)
        :return: If that property was retrieved via this method. Returns None
            if the key was not in the retrieved values.
        :rtype: various (the value), bool (if the default value is used)
        """
        if 'general' not in self._cache:
            pywikibot.debug('general siteinfo not loaded yet.')
            force = True
            props = ['namespaces', 'namespacealiases']
        else:
            force = Siteinfo._is_expired(self._cache['general'][1], expiry)
            props = []
        if force:
            props = [prop for prop in props if prop not in self._cache]
            if props:
                pywikibot.debug(
                    "Load siteinfo properties '{}' along with 'general'"
                    .format("', '".join(props)))
            props.append('general')
            default_info = self._get_siteinfo(props, expiry)
            for prop in props:
                self._cache[prop] = default_info[prop]
            if key in default_info:
                return default_info[key]

        if key in self._cache['general'][0]:
            return self._cache['general'][0][key], self._cache['general']

        return None

    def __getitem__(self, key: str):
        """Return a siteinfo property, caching and not forcing it."""
        return self.get(key, False)  # caches and doesn't force it

    def get(
        self,
        key: str,
        get_default: bool = True,
        cache: bool = True,
        expiry: datetime.datetime | float | bool = False
    ) -> Any:
        """Return a siteinfo property.

        It will never throw an APIError if it only stated, that the siteinfo
        property doesn't exist. Instead it will use the default value.

        .. seealso:: :py:obj:`_get_siteinfo`

        :param key: The name of the siteinfo property.
        :param get_default: Whether to throw an KeyError if the key is invalid.
        :param cache: Caches the result internally so that future accesses via
            this method won't query the server.
        :param expiry: If the cache is older than the expiry it ignores the
            cache and queries the server to get the newest value.
        :return: The gathered property
        :raises KeyError: If the key is not a valid siteinfo property and the
            get_default option is set to False.
        """
        # If expiry is True, convert it to 0 to be coherent with
        # _get_siteinfo() and _get_general() docstring.
        if expiry is True:
            expiry = 0
        # If expiry is a float or int convert to timedelta
        # Note: bool is an instance of int
        if isinstance(expiry, float) or type(expiry) is int:
            expiry = datetime.timedelta(expiry)

        # expire = 0 (or timedelta(0)) are always expired and their bool is
        # False, so skip them EXCEPT if it's literally False, then they expire
        # never.
        if expiry and expiry is not True or expiry is False:
            try:
                cached = self._get_cached(key)
            except KeyError:
                pass
            else:  # cached value available
                # is a default value, but isn't accepted
                if not cached[1] and not get_default:
                    raise KeyError(key)
                if not Siteinfo._is_expired(cached[1], expiry):
                    return copy.deepcopy(cached[0])

        preloaded = self._get_general(key, expiry)
        if not preloaded:
            preloaded = self._get_siteinfo(key, expiry)[key]
        else:
            cache = False

        if not preloaded[1] and not get_default:
            raise KeyError(key)

        if cache:
            self._cache[key] = preloaded

        return copy.deepcopy(preloaded[0])

    def _get_cached(self, key: str):
        """Return the cached value or a KeyError exception if not cached."""
        if 'general' in self._cache:
            if key in self._cache['general'][0]:
                return (self._cache['general'][0][key],
                        self._cache['general'][1])
            return self._cache[key]
        raise KeyError(key)

    def is_cached(self, key: str) -> bool:
        """Return whether the value is cached.

        .. versionadded:: 7.1
        """
        try:
            self._get_cached(key)
        except KeyError:
            return False

        return True

    def __contains__(self, key: object) -> bool:
        """Check whether the given key is present in the Siteinfo container.

        This method implements the Container protocol and allows usage
        like `key in container`.Only string keys are valid. Non-string
        keys always return False.

        .. versionchanged:: 7.1
           Previous implementation only checked for cached keys.

        :param key: The key to check for presence. Should be a string.
        :return: True if the key exists in the container, False otherwise.

        :meta public:
        """
        if isinstance(key, str):
            with suppress(KeyError):
                self[key]
                return True

        return False

    def is_recognised(self, key: str) -> bool | None:
        """Return if 'key' is a valid property name.

        'None' if not cached.
        """
        time = self.get_requested_time(key)
        return None if time is None else bool(time)

    def get_requested_time(self, key: str):
        """Return when 'key' was successfully requested from the server.

        If the property is actually in the siprop 'general' it returns the
        last request from the 'general' siprop.

        :param key: The siprop value or a property of 'general'.
        :return: The last time the siprop of 'key' was requested.
        :rtype: None (never), False (default),
            :py:obj:`datetime.datetime` (cached)
        """
        with suppress(KeyError):
            return self._get_cached(key)[1]

        return None



"""Character based helper functions (not wiki-dependent)."""
#
# (C) Pywikibot team, 2015-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import re
from contextlib import suppress
from typing import cast
from urllib.parse import unquote

from pywikibot.backports import Iterable
from pywikibot.tools._unidata import _category_cf


# This is a set of all invisible characters
# At the moment we've only added the characters from the Cf category
_invisible_chars = _category_cf

INVISIBLE_REGEX = re.compile(f"[{''.join(_invisible_chars)}]")


def contains_invisible(text):
    """Return True if the text contain any of the invisible characters."""
    return any(char in _invisible_chars for char in text)


def replace_invisible(text):
    """Replace invisible characters by '<codepoint>'."""
    def replace(match) -> str:
        codepoint = ord(match.group())
        return f'<{codepoint:x}>'

    return INVISIBLE_REGEX.sub(replace, text)


def string_to_ascii_html(string: str) -> str:
    """Convert unicode chars of str to HTML entities if chars are not ASCII.

    **Example:**

    >>> string_to_ascii_html('Python')
    'Python'
    >>> string_to_ascii_html("Pywikibot's API")
    "Pywikibot's API"
    >>> string_to_ascii_html('Eetße Joohunndot füür Kreůßtůß')
    'Eet&#223;e Joohunndot f&#252;&#252;r Kre&#367;&#223;t&#367;&#223;'

    :param string: String to update
    """
    html = []
    for c in string:
        cord = ord(c)
        if 31 < cord < 127:
            html.append(c)
        else:
            html.append(f'&#{cord};')
    return ''.join(html)


def string2html(string: str, encoding: str) -> str:
    """Convert unicode string to requested HTML encoding.

    Attempt to encode the string into the desired format; if that work
    return it unchanged. Otherwise encode the non-ASCII characters into
    HTML &#; entities.

    **Example:**

    >>> string2html('Referências', 'utf-8')
    'Referências'
    >>> string2html('Referências', 'ascii')
    'Refer&#234;ncias'
    >>> string2html('脚注', 'euc_jp')
    '脚注'
    >>> string2html('脚注', 'iso-8859-1')
    '&#33050;&#27880;'

    :param string: String to update
    :param encoding: Encoding to use
    """
    with suppress(UnicodeError):
        string.encode(encoding)
        return string

    return string_to_ascii_html(string)


def url2string(title: str,
               encodings: str | Iterable[str] = 'utf-8') -> str:
    """Convert URL-encoded text to unicode using several encoding.

    Uses the first encoding that doesn't cause an error. Raises the
    first exception if all encodings fails.

    For a single *encodings* string this function is equivalent to
    :samp:`urllib.parse.unquote(title, encodings, errors='strict')`

    .. versionchanged:: 8.4
       Ignore *LookupError* and try other encodings.

    .. seealso:: :python:`urllib.parse.unquote
       <library/urllib.parse.html#urllib.parse.unquote>`

    **Example:**

    >>> url2string('abc%20def')
    'abc def'
    >>> url2string('/El%20Ni%C3%B1o/')
    '/El Niño/'
    >>> url2string('/El%20Ni%C3%B1o/', 'ascii')
    Traceback (most recent call last):
    ...
    UnicodeDecodeError: 'ascii' codec can't decode byte 0xc3 in position 6:...
    >>> url2string('/El%20Ni%C3%B1o/', ['ascii', 'utf-8'])
    '/El Niño/'

    :param title: URL-encoded character data to convert
    :param encodings: Encodings to attempt to use during conversion.

    :raise UnicodeError: Could not convert using any encoding.
    :raise LookupError: unknown encoding
    """
    if isinstance(encodings, str):
        return unquote(title, encodings, errors='strict')

    first_exception: BaseException | None = None
    for enc in encodings:
        try:
            return unquote(title, enc, errors='strict')
        except (UnicodeError, LookupError) as e:
            if first_exception is None:
                first_exception = e

    # Couldn't convert, raise the first exception
    raise cast(BaseException, first_exception)



"""Family module for Vikidia."""
#
# (C) Pywikibot team, 2010-2024
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from pywikibot import family
from pywikibot.tools import classproperty


class Family(family.SubdomainFamily):

    """Family class for Vikidia."""

    name = 'vikidia'
    domain = 'vikidia.org'

    codes = {
        'ar', 'ca', 'de', 'el', 'en', 'es', 'eu', 'fr', 'hy', 'it', 'nl', 'oc',
        'pt', 'ru', 'scn',
    }

    # Sites we want to edit but not count as real languages
    test_codes = ['central', 'test']

    @classproperty
    def domains(cls):
        """List of domains used by Vikidia family."""
        return [
            cls.domain,
            'wikikids.nl'  # nl
        ]

    @classproperty
    def langs(cls):
        """Property listing family languages."""
        cls.langs = super().langs
        cls.langs['nl'] = cls.domains[1]
        return cls.langs

    def scriptpath(self, code):
        """Return the script path for this family."""
        if code == 'nl':
            return ''
        return super().scriptpath(code)



"""Title translate module."""
#
# (C) Pywikibot team, 2003-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import pywikibot
from pywikibot import config, date


def translate(
    page=None,
    hints: list[str] | None = None,
    auto: bool = True,
    removebrackets: bool = False,
    site=None
) -> list[pywikibot.Link]:
    """Return a list of links to pages on other sites based on hints.

    Entries for single page titles list those pages. Page titles for
    entries such as "all:" or "xyz:" or "20:" are first built from the
    page title of 'page' and then listed.

    .. versionchanged:: 9.6
       Raise ``RuntimeError`` instead of ``AssertionError`` if neither
       *page* nor *site* parameter is given.

    :param auto: If  true, known year and date page titles are
        autotranslated to all known target languages and inserted into
        the list.
    :param removebrackets: If True, a trailing pair of brackets and the
        text between them is removed from the page title.
    :raises RuntimeError: Either page or site parameter must be given.
    """
    if not page and not site:
        raise RuntimeError(
            'Either page or site parameter must be given with translate()')

    site = site or page.site
    result = set()

    if hints is None:
        hints = []

    for h in hints:
        # argument may be given as -hint:xy where xy is a language code
        code, _, newname = h.partition(':')
        if not newname:
            # if given as -hint:xy or -hint:xy:, assume that there should
            # be a page in language xy with the same title as the page
            # we're currently working on ...
            if page is None:
                continue
            newname = page.title(with_ns=False,
                                 without_brackets=removebrackets)
        if code.isdigit():
            codes = site.family.languages_by_size[:int(code)]
        elif code == 'all':
            codes = list(site.family.codes)
        else:
            codes = site.family.language_groups.get(code, code.split(','))

        for newcode in codes:
            if newcode in site.codes:
                if newcode != site.code:
                    ns = page.namespace() if page else 0
                    link = pywikibot.Link(newname,
                                          site.getSite(code=newcode),
                                          default_namespace=ns)
                    result.add(link)
            elif config.verbose_output:
                pywikibot.info(f'Ignoring unknown language code {newcode}')

    # Autotranslate dates into all other languages, the rest will come from
    # existing interwiki links.
    if auto and page:
        # search inside all dictionaries for this link
        dict_name, value = page.autoFormat()
        if dict_name:
            pywikibot.info(f'TitleTranslate: {page.title()} was recognized as '
                           f'{dict_name} with value {value}')
            for entry_lang, entry in date.formats[dict_name].items():
                if entry_lang not in site.codes:
                    continue

                if entry_lang != page.site.lang:
                    link = pywikibot.Link(entry(value),
                                          site.getSite(entry_lang))
                    result.add(link)

    return list(result)



"""Folder containing i18n translations bundles.

This file exists to help pkgutil find the i18n data. There is no actual
python code in this package.
"""
from __future__ import annotations


#
# (C) Pywikibot team, 2011-2022
#
# Distributed under the terms of the MIT license.
#
# Translations by translators are licensed CC BY 3.0
#



"""Objects representing MediaWiki log entries."""
#
# (C) Pywikibot team, 2007-2024
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import datetime
from collections import UserDict
from typing import Any

import pywikibot
from pywikibot.exceptions import Error, HiddenKeyError
from pywikibot.tools import cached


class LogEntry(UserDict):

    """Generic log entry.

    LogEntry parameters may be retrieved by the corresponding method or
    the LogEntry key. The following statements are equivalent:

    action = logentry.action() action = logentry['action'] action =
    logentry.data['action']
    """

    # Log type expected. None for every type, or one of the (letype) str :
    # block/patrol/etc...
    # Overridden in subclasses.
    _expected_type: str | None = None

    def __init__(self, apidata: dict[str, Any],
                 site: pywikibot.site.BaseSite) -> None:
        """Initialize object from a logevent dict returned by MW API."""
        super().__init__(apidata)
        self.site = site
        expected_type = self._expected_type
        if expected_type is not None and expected_type != self.type():
            raise Error(f'Wrong log type! Expecting {expected_type}, received '
                        f'{self.type()} instead.')

    def __missing__(self, key: str) -> None:
        """Debug when the key is missing.

        HiddenKeyError is raised when the user does not have permission.
        KeyError is raised otherwise.

        It also logs debugging information when a key is missing.
        """
        pywikibot.debug(f'API log entry received:\n{self!r}')
        hidden = {
            'actionhidden': [
                'action', 'logpage', 'ns', 'pageid', 'params', 'title',
            ],
            'commenthidden': ['comment'],
            'userhidden': ['user'],
        }
        for hidden_key, hidden_types in hidden.items():
            if hidden_key in self and key in hidden_types:
                raise HiddenKeyError(
                    f'Log entry ({self["type"]}) has a hidden {key!r} key and'
                    " you don't have permission to view it due to "
                    f'{hidden_key!r}'
                )

        raise KeyError(f"Log entry ({self['type']}) has no {key!r} key")

    def __repr__(self) -> str:
        """Return a string representation of LogEntry object."""
        return (f'<{type(self).__name__}({self.site.sitename}, '
                f'logid={self.logid()})>')

    def __hash__(self) -> int:
        """Combine site and logid as the hash."""
        return self.logid() ^ hash(self.site)

    def __eq__(self, other: Any) -> bool:
        """Compare if self is equal to other."""
        if not isinstance(other, LogEntry):
            pywikibot.debug(f"'{type(self).__name__}' cannot be compared with "
                            f"'{type(other).__name__}'")
            return False

        return self.logid() == other.logid() and self.site == other.site

    def __getattr__(self, item: str) -> Any:
        """Return several items from dict used as methods."""
        if item in ('action', 'comment', 'logid', 'ns', 'pageid', 'type',
                    'user'):  # TODO use specific User class for 'user'?
            return lambda: self[item]

        return super().__getattribute__(item)

    @property
    def params(self) -> dict[str, Any]:
        """Additional data for some log entry types.

        .. versionadded:: 9.4
           private *_param* attribute became a public property
        """
        return self.get('params', {})

    @cached
    def page(self) -> int | pywikibot.page.Page:
        """Page on which action was performed.

        :return: page on action was performed
        """
        return pywikibot.Page(self.site, self['title'])

    @cached
    def timestamp(self) -> pywikibot.Timestamp:
        """Timestamp object corresponding to event timestamp."""
        return pywikibot.Timestamp.fromISOformat(self['timestamp'])


class OtherLogEntry(LogEntry):

    """A log entry class for unspecified log events."""


class UserTargetLogEntry(LogEntry):

    """A log entry whose target is a user page."""

    @cached
    def page(self) -> pywikibot.page.User:
        """Return the target user.

        This returns a User object instead of the Page object returned
        by the superclass method.

        :return: target user
        """
        return pywikibot.User(self.site, self['title'])


class BlockEntry(LogEntry):

    """Block or unblock log entry.

    It might contain a block or unblock depending on the action. The
    duration, expiry and flags are not available on unblock log entries.
    """

    _expected_type = 'block'

    def __init__(self, apidata: dict[str, Any],
                 site: pywikibot.site.BaseSite) -> None:
        """Initializer."""
        super().__init__(apidata, site)
        # When an autoblock is removed, the "title" field is not a page title
        # See bug T19781
        pos = self.get('title', '').find('#')
        self.isAutoblockRemoval = pos > 0
        if self.isAutoblockRemoval:
            self._blockid = int(self['title'][pos + 1:])

    def page(self) -> int | pywikibot.page.Page:
        """Return the blocked account or IP.

        :return: the Page object of username or IP if this block action
            targets a username or IP, or the blockid if this log
            reflects the removal of an autoblock
        """
        # TODO what for IP ranges ?
        if self.isAutoblockRemoval:
            return self._blockid

        return super().page()

    @cached
    def flags(self) -> list[str]:
        """Return a list of (str) flags associated with the block entry.

        It raises an Error if the entry is an unblocking log entry.

        :return: list of flags strings
        """
        if self.action() == 'unblock':
            return []

        return self.params.get('flags', [])

    @cached
    def duration(self) -> datetime.timedelta | None:
        """Return a datetime.timedelta representing the block duration.

        :return: datetime.timedelta, or None if block is indefinite.
        """
        # Doing the difference is easier than parsing the string
        return (self.expiry() - self.timestamp()
                if self.expiry() is not None else None)

    @cached
    def expiry(self) -> pywikibot.Timestamp | None:
        """Return a Timestamp representing the block expiry date."""
        details = self.params.get('expiry')
        return pywikibot.Timestamp.fromISOformat(details) if details else None


class RightsEntry(LogEntry):

    """Rights log entry."""

    _expected_type = 'rights'

    @property
    def oldgroups(self) -> list[str]:
        """Return old rights groups.

        .. versionchanged:: 7.5
           No longer raise KeyError if `oldgroups` does not exists or
           LogEntry has no additional data e.g. due to hidden data and
           insufficient rights.
        """
        return self.params.get('oldgroups', [])

    @property
    def newgroups(self) -> list[str]:
        """Return new rights groups.

        .. versionchanged:: 7.5
           No longer raise KeyError if `oldgroups` does not exists or
           LogEntry has no additional data e.g. due to hidden data and
           insufficient rights.
        """
        return self.params.get('newgroups', [])


class UploadEntry(LogEntry):

    """Upload log entry."""

    _expected_type = 'upload'

    @cached
    def page(self) -> pywikibot.page.FilePage:
        """Return FilePage on which action was performed."""
        return pywikibot.FilePage(self.site, self['title'])


class MoveEntry(LogEntry):

    """Move log entry."""

    _expected_type = 'move'

    @property
    def target_ns(self) -> pywikibot.site._namespace.Namespace:
        """Return namespace object of target page."""
        return self.site.namespaces[self.params['target_ns']]

    @property
    def target_title(self) -> str:
        """Return the target title."""
        return self.params['target_title']

    @property
    @cached
    def target_page(self) -> pywikibot.page.Page:
        """Return target page object."""
        return pywikibot.Page(self.site, self.target_title)

    def suppressedredirect(self) -> bool:
        """Return True if no redirect was created during the move."""
        # Introduced in MW r47901
        return 'suppressedredirect' in self.params


class PatrolEntry(LogEntry):

    """Patrol log entry."""

    _expected_type = 'patrol'

    @property
    def current_id(self) -> int:
        """Return the current id."""
        return int(self.params['curid'])

    @property
    def previous_id(self) -> int:
        """Return the previous id."""
        return int(self.params['previd'])

    @property
    def auto(self) -> bool:
        """Return auto patrolled."""
        return 'auto' in self.params and self.params['auto'] != 0


class LogEntryFactory:

    """LogEntry Factory.

    Only available method is create()
    """

    _logtypes = {
        'block': BlockEntry,
        'rights': RightsEntry,
        'upload': UploadEntry,
        'move': MoveEntry,
        'patrol': PatrolEntry,
    }

    def __init__(self, site: pywikibot.site.BaseSite,
                 logtype: str | None = None) -> None:
        """Initializer.

        :param site: The site on which the log entries are created.
        :param logtype: The log type of the log entries, if known in
            advance. If None, the Factory will fetch the log entry from
            the data to create each object.
        """
        self._site = site
        if logtype is None:
            self._creator = self._create_from_data
        else:
            # Bind a Class object to self._creator:
            # When called, it will initialize a new object of that class
            logclass = self.get_valid_entry_class(logtype)
            self._creator = lambda data: logclass(data, self._site)

    def create(self, logdata: dict[str, Any]) -> LogEntry:
        """Instantiate the LogEntry object representing logdata.

        :param logdata: <item> returned by the api
        :return: LogEntry object representing logdata
        """
        return self._creator(logdata)

    def get_valid_entry_class(self, logtype: str) -> LogEntry:
        """Return the class corresponding to the @logtype string parameter.

        :return: specified subclass of LogEntry
        :raise KeyError: logtype is not valid
        """
        if logtype not in self._site.logtypes:
            raise KeyError(f'{logtype} is not a valid logtype')

        return LogEntryFactory.get_entry_class(logtype)

    @classmethod
    def get_entry_class(cls, logtype: str) -> LogEntry:
        """Return the class corresponding to the @logtype string parameter.

        :return: specified subclass of LogEntry

        .. note:: this class method cannot verify whether the given logtype
           already exits for a given site; to verify use Site.logtypes
           or use the get_valid_entry_class instance method instead.
        """
        if logtype not in cls._logtypes:
            if logtype is None:
                cls._logtypes[logtype] = OtherLogEntry
            else:
                if logtype in ('newusers', 'thanks'):
                    bases = (UserTargetLogEntry, OtherLogEntry)
                else:
                    bases = (OtherLogEntry,)
                cls._logtypes[logtype] = type(
                    f'{logtype.capitalize()}Entry',
                    bases,
                    {
                        '__doc__': f'{logtype.capitalize()} log entry',
                        '_expected_type': logtype,
                    },
                )
        return cls._logtypes[logtype]

    def _create_from_data(self, logdata: dict[str, Any]) -> LogEntry:
        """Check for logtype from data, and creates the correct LogEntry.

        :param logdata: log entry data
        """
        try:
            logtype = logdata['type']
        except KeyError:
            pywikibot.debug(f'API log entry received:\n{logdata}')
            raise Error("Log entry has no 'type' key")

        return LogEntryFactory.get_entry_class(logtype)(logdata, self._site)


# For backward compatibility
ProtectEntry = LogEntryFactory.get_entry_class('protect')
DeleteEntry = LogEntryFactory.get_entry_class('delete')
ImportEntry = LogEntryFactory.get_entry_class('import')
NewUsersEntry = LogEntryFactory.get_entry_class('newusers')
ThanksEntry = LogEntryFactory.get_entry_class('thanks')



"""Objects representing MediaWiki File pages.

This module includes objects:

* FilePage: A subclass of Page representing a file description page
* FileInfo: a structure holding imageinfo of latest revision of FilePage
"""
#
# (C) Pywikibot team, 2008-2024
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from http import HTTPStatus
from os import PathLike
from pathlib import Path
from urllib.parse import urlparse

import pywikibot
from pywikibot.backports import Iterable
from pywikibot.comms import http
from pywikibot.exceptions import NoPageError
from pywikibot.page._page import Page
from pywikibot.tools import compute_file_hash


__all__ = (
    'FileInfo',
    'FilePage',
)


class FilePage(Page):

    """A subclass of Page representing a file description page.

    Supports the same interface as Page except *ns*; some added methods.
    """

    def __init__(self, source, title: str = '', *,
                 ignore_extension: bool = False) -> None:
        """Initializer.

        .. versionchanged:: 8.4
           Check for valid extensions.
        .. versionchanged:: 9.3
           Added the optional *ignore_extension* parameter.
        .. versionchanged:: 9.6
           Show a warning if *ignore_extension* was set and the
           extension is invalid.
        .. seealso::
           :meth:`Site.file_extensions
           <pywikibot.site._apisite.APISite.file_extensions>`

        :param source: the source of the page
        :type source: pywikibot.page.BaseLink (or subclass),
            pywikibot.page.Page (or subclass), or pywikibot.page.Site
        :param title: normalized title of the page; required if source is a
            Site, ignored otherwise
        :param ignore_extension: prevent extension check
        :raises ValueError: Either the title is not in the file
            namespace or does not have a valid extension and
            *ignore_extension* was not set.
        """
        self._file_revisions = {}  # dictionary to cache File history.
        super().__init__(source, title, 6)
        if self.namespace() != 6:
            raise ValueError(f"'{self.title()}' is not in the file namespace!")

        title = self.title(with_ns=False, with_section=False)
        _, sep, extension = title.rpartition('.')
        if not sep or extension.lower() not in self.site.file_extensions:
            msg = (f'{title!r} does not have a valid extension\n'
                   f'({", ".join(self.site.file_extensions)}).')
            if not ignore_extension:
                raise ValueError(msg)

            pywikibot.warning(msg)

    def _load_file_revisions(self, imageinfo) -> None:
        """Save a file revision of FilePage (a FileInfo object) in local cache.

        Metadata shall be added lazily to the revision already present
        in cache.
        """
        for file_rev in imageinfo:
            # filemissing in API response indicates most fields are missing
            # see https://gerrit.wikimedia.org/r/c/mediawiki/core/+/533482/
            if 'filemissing' in file_rev:
                pywikibot.warning(
                    f"File '{self.title()}' contains missing revisions")
                continue

            ts_key = pywikibot.Timestamp.fromISOformat(file_rev['timestamp'])
            file_revision = self._file_revisions.setdefault(
                ts_key, FileInfo(file_rev, self))

            # add new imageinfo attributes since last request.
            file_revision.update(file_rev)

    @property
    def latest_file_info(self):
        """Retrieve and store information of latest Image rev. of FilePage.

        At the same time, the whole history of Image is fetched and
        cached in self._file_revisions

        :return: instance of FileInfo()
        """
        if not self._file_revisions:
            self.site.loadimageinfo(self, history=True)
        latest_ts = max(self._file_revisions)
        return self._file_revisions[latest_ts]

    @property
    def oldest_file_info(self):
        """Retrieve and store information of oldest Image rev. of FilePage.

        At the same time, the whole history of Image is fetched and
        cached in self._file_revisions

        :return: instance of FileInfo()
        """
        if not self._file_revisions:
            self.site.loadimageinfo(self, history=True)
        oldest_ts = min(self._file_revisions)
        return self._file_revisions[oldest_ts]

    def get_file_info(self, ts) -> dict:
        """Retrieve and store information of a specific Image rev. of FilePage.

        This function will load also metadata.
        It is also used as a helper in FileInfo to load metadata lazily.

        .. versionadded:: 8.6

        :param ts: timestamp of the Image rev. to retrieve

        :return: instance of FileInfo()
        """
        self.site.loadimageinfo(self, history=False, timestamp=ts)
        return self._file_revisions[ts]

    def get_file_history(self) -> dict:
        """Return the file's version history.

        :return: dictionary with:
            key: timestamp of the entry
            value: instance of FileInfo()
        """
        if not self._file_revisions:
            self.site.loadimageinfo(self, history=True)
        return self._file_revisions

    def getImagePageHtml(self) -> str:  # noqa: N802
        """Download the file page, and return the HTML, as a string.

        Caches the HTML code, so that if you run this method twice on
        the same FilePage object, the page will only be downloaded once.
        """
        if not hasattr(self, '_imagePageHtml'):
            path = (f'{self.site.scriptpath()}/index.php?'
                    f'title={self.title(as_url=True)}')
            self._imagePageHtml = http.request(self.site, path).text
        return self._imagePageHtml

    def get_file_url(self,
                     url_width: int | None = None,
                     url_height: int | None = None,
                     url_param: str | None = None) -> str:
        """Return the url or the thumburl of the file described on this page.

        Fetch the information if not available.

        Once retrieved, file information will also be accessible as
        :attr:`latest_file_info` attributes, named as in :api:`Imageinfo`.
        If *url_width*, *url_height* or *url_param* is given, additional
        properties ``thumbwidth``, ``thumbheight``, ``thumburl`` and
        ``responsiveUrls`` are provided.

        .. note:: Parameters validation and error handling left to the
           API call.
        .. seealso::

           * :meth:`APISite.loadimageinfo()
             <pywikibot.site._apisite.APISite.loadimageinfo>`
           * :api:`Imageinfo`

        :param url_width: get info for a thumbnail with given width
        :param url_height: get info for a thumbnail with given height
        :param url_param:  get info for a thumbnail with given param
        :return: latest file url or thumburl
        """
        # Plain url is requested.
        if url_width is None and url_height is None and url_param is None:
            return self.latest_file_info.url

        # Thumburl is requested.
        self.site.loadimageinfo(self, history=not self._file_revisions,
                                url_width=url_width, url_height=url_height,
                                url_param=url_param)
        return self.latest_file_info.thumburl

    def file_is_shared(self) -> bool:
        """Check if the file is stored on any known shared repository.

        .. versionchanged:: 7.0
           return False if file does not exist on shared image repository
           instead raising NoPageError.
        """
        # as of now, the only known repositories are commons and wikitravel
        # TODO: put the URLs to family file
        if not self.site.has_image_repository:
            return False

        try:
            info = self.latest_file_info
        except NoPageError:
            return False

        if 'wikitravel_shared' in self.site.shared_image_repository():
            return info.url.startswith('https://wikitravel.org/upload/shared/')

        # default to commons
        return info.url.startswith(
            'https://upload.wikimedia.org/wikipedia/commons/')

    def getFileVersionHistoryTable(self) -> str:  # noqa: N802
        """Return the version history in the form of a wiki table."""
        lines = []
        for info in self.get_file_history().values():
            dimension = '{width}×{height} px ({size} bytes)'.format(
                **info.__dict__)
            lines.append('| {timestamp} || {user} || {dimension} |'
                         '| <nowiki>{comment}</nowiki>'
                         .format(dimension=dimension, **info.__dict__))
        return ('{| class="wikitable"\n'
                '! {{int:filehist-datetime}} || {{int:filehist-user}} |'
                '| {{int:filehist-dimensions}} || {{int:filehist-comment}}\n'
                '|-\n%s\n|}\n' % '\n|-\n'.join(lines))

    def using_pages(self, **kwargs):
        """Yield Pages on which the file is displayed.

        For parameters refer
        :meth:`APISite.imageusage()
        <pywikibot.site._generators.GeneratorsMixin.imageusage>`

        Usage example:

        >>> site = pywikibot.Site('wikipedia:test')
        >>> file = pywikibot.FilePage(site, 'Pywikibot MW gear icon.svg')
        >>> used = list(file.using_pages(total=10))
        >>> len(used)
        2
        >>> used[0].title()
        'Pywikibot'

        .. seealso:: :meth:`globalusage`
        .. versionchanged:: 7.2
           all parameters from :meth:`APISite.imageusage()
           <pywikibot.site._generators.GeneratorsMixin.imageusage>`
           are available.
        .. versionchanged:: 7.4
           renamed from :meth:`usingPages`.
        """
        return self.site.imageusage(self, **kwargs)

    @property
    def file_is_used(self) -> bool:
        """Check whether the file is used at this site.

        .. versionadded:: 7.1
        """
        return bool(list(self.using_pages(total=1)))

    def upload(self, source: str, **kwargs) -> bool:
        """Upload this file to the wiki.

        keyword arguments are from site.upload() method.

        :param source: Path or URL to the file to be uploaded.

        :keyword comment: Edit summary; if this is not provided, then
            filepage.text will be used. An empty summary is not
            permitted. This may also serve as the initial page text (see
            below).
        :keyword text: Initial page text; if this is not set, then
            filepage.text will be used, or comment.
        :keyword watch: If true, add filepage to the bot user's
            watchlist
        :keyword ignore_warnings: It may be a static boolean, a callable
            returning a boolean or an iterable. The callable gets a list
            of UploadError instances and the iterable should contain the
            warning codes for which an equivalent callable would return
            True if all UploadError codes are in that list. If the
            result is False it'll not continue uploading the file and
            otherwise disable any warning and reattempt to upload the
            file.

            .. note:: NOTE: If report_success is True or None it'll
               raise an UploadError exception if the static boolean is
               False.
        :type ignore_warnings: bool or callable or iterable of str
        :keyword chunk_size: The chunk size in bytesfor chunked
            uploading (see :api:`Upload#Chunked_uploading`). It will
            only upload in chunks, if the chunk size is positive but
            lower than the file size.
        :type chunk_size: int
        :keyword report_success: If the upload was successful it'll
            print a success message and if ignore_warnings is set to
            False it'll raise an UploadError if a warning occurred. If
            it's None (default) it'll be True if ignore_warnings is a
            bool and False otherwise. If it's True or None
            ignore_warnings must be a bool.
        :return: It returns True if the upload was successful and False
            otherwise.
        """
        filename = url = None
        if '://' in source:
            url = source
        else:
            filename = source
        return self.site.upload(self, source_filename=filename, source_url=url,
                                **kwargs)

    def download(self,
                 filename: str | PathLike | Iterable[str] | None = None,
                 chunk_size: int = 100 * 1024,
                 revision: FileInfo | None = None, *,
                 url_width: int | None = None,
                 url_height: int | None = None,
                 url_param: str | None = None) -> bool:
        """Download to filename file of FilePage.

        **Usage examples:**

        Download an image:

        >>> site = pywikibot.Site('wikipedia:test')
        >>> file = pywikibot.FilePage(site, 'Pywikibot MW gear icon.svg')
        >>> file.download()
        True

        Pywikibot_MW_gear_icon.svg was downloaded.

        Download a thumbnail:

        >>> file.download(url_param='120px')
        True

        The suffix has changed and Pywikibot_MW_gear_icon.png was
        downloaded.

        .. versionadded:: 8.2
           *url_width*, *url_height* and *url_param* parameters.
        .. versionchanged:: 8.2
           *filename* argument may be also a path-like object or an
           iterable of path segments.
        .. note:: filename suffix is adjusted if target url's suffix is
           different which may be the case if a thumbnail is loaded.
        .. warning:: If a file already exists, it will be overridden
           without further notes.
        .. seealso:: :api:`Imageinfo` for new parameters

        :param filename: filename where to save file. If ``None``,
            ``self.title(as_filename=True, with_ns=False)`` will be used.
            If an Iterable is specified the items will be used as path
            segments. To specify the user directory path you have to use
            either ``~`` or ``~user`` as first path segment e.g. ``~/foo``
            or ``('~', 'foo')`` as filename. If only the user directory
            specifier is given, the title is used as filename like for
            None. If the suffix is missing or different from url (which
            can happen if a *url_width*, *url_height* or *url_param*
            argument is given), the file suffix is adjusted.
        :param chunk_size: the size of each chunk to be received and
            written to file.
        :param revision: file revision to download. If None
            :attr:`latest_file_info` will be used; otherwise provided
            revision will be used.
        :param url_width: download thumbnail with given width
        :param url_height: download thumbnail with given height
        :param url_param:  download thumbnail with given param
        :return: True if download is successful, False otherwise.
        :raise IOError: if filename cannot be written for any reason.
        """
        if not filename:
            path = Path()
        elif isinstance(filename, (str, PathLike)):
            path = Path(filename)
        else:
            path = Path(*filename)

        if path.stem in ('', '~', '~user'):
            path = path / self.title(as_filename=True, with_ns=False)

        thumb = bool(url_width or url_height or url_param)
        if thumb or revision is None:
            url = self.get_file_url(url_width, url_height, url_param)
            revision = self.latest_file_info
        else:
            url = revision.url

        # adjust suffix
        path = path.with_suffix(Path(urlparse(url).path).suffix)
        # adjust user path
        path = path.expanduser()
        req = http.fetch(url, stream=True)
        if req.status_code == HTTPStatus.OK:
            with open(path, 'wb') as f:
                for chunk in req.iter_content(chunk_size):
                    f.write(chunk)

            return thumb or compute_file_hash(path) == revision.sha1

        pywikibot.warning(
            f'Unsuccessful request ({req.status_code}): {req.url}')
        return False

    def globalusage(self, total=None):
        """Iterate all global usage for this page.

        .. seealso:: :meth:`using_pages`

        :param total: iterate no more than this number of pages in total
        :return: a generator that yields Pages also on sites different from
            self.site.
        :rtype: generator
        """
        return self.site.globalusage(self, total=total)

    def data_item(self):
        """Function to get the associated Wikibase item of the file.

        If WikibaseMediaInfo extension is available (e.g., on Commons),
        the method returns the associated mediainfo entity. Otherwise,
        it falls back to the behavior of :meth:`BasePage.data_item`.

        .. versionadded:: 6.5

        :rtype: pywikibot.page.WikibaseEntity
        """
        if self.site.has_extension('WikibaseMediaInfo'):
            if not hasattr(self, '_item'):
                self._item = pywikibot.MediaInfo(self.site)
                self._item._file = self
            return self._item

        return super().data_item()


class FileInfo:

    """A structure holding imageinfo of latest rev. of FilePage.

    All keys of API imageinfo dictionary are mapped to FileInfo attributes.
    Attributes can be retrieved both as self['key'] or self.key.

    Following attributes will be returned:
        - timestamp, user, comment, url, size, sha1, mime, metadata (lazily)
        - archivename (not for latest revision)

    see :meth:`Site.loadimageinfo()
    <pywikibot.site._apisite.APISite.loadimageinfo>` for details.

    .. note:: timestamp will be casted to :func:`pywikibot.Timestamp`.

    .. versionchanged:: 7.7
       raises KeyError instead of AttributeError if FileInfo is used as
       Mapping.
    .. versionchanged:: 8.6
       Metadata are loaded lazily.
       Added *filepage* parameter.
    """

    def __init__(self, file_revision, filepage) -> None:
        """Initiate the class using the dict from ``APISite.loadimageinfo``."""
        self.filepage = filepage
        self._metadata = None
        self.update(file_revision)

    def update(self, file_revision) -> None:
        """Update FileInfo with new values.

        .. versionadded:: 8.6
        """
        for k, v in file_revision.items():
            if k == 'timestamp':
                v = pywikibot.Timestamp.fromISOformat(v)
            setattr(self, k, v)

    def __getitem__(self, key):
        """Give access to class values by key."""
        try:
            result = getattr(self, key)
        except AttributeError as e:
            raise KeyError(str(e).replace('attribute', 'key')) from None
        return result

    def __repr__(self) -> str:
        """Return a more complete string representation."""
        return repr(self.__dict__)

    def __eq__(self, other) -> bool:
        """Test if two FileInfo objects are equal."""
        return self.__dict__ == other.__dict__

    @property
    def metadata(self):
        """Return metadata.

        .. versionadded:: 8.6
        """
        if self._metadata is None:
            self.filepage.get_file_info(self.timestamp)
        return self._metadata

    @metadata.setter
    def metadata(self, value) -> None:
        """Set metadata.

        .. versionadded:: 8.6
        """
        self._metadata = value



#!/usr/bin/env python3
"""Script to determine the Pywikibot version (tag, revision and date).

The following option is supported:

-nouser  do not print usernames; otherwise they are printed for each
         registered family

.. versionchanged:: 7.0
   version script was moved to the framework scripts folder.
.. versionadded:: 9.1.2
   the *-nouser* option was added.
.. versionchanged:: 10.6
   The User-Agent string is now printed for the default site. To print
   it for another site, call the ``pwb`` wrapper with the global option,
   e.g.:

       pwb -site:wikipedia:test version

   .. note::
      The shown UA reflects the default config settings. It might differ
      if a user-agent is passed via the *headers* parameter to
      :func:`comms.http.request`, :func:`comms.http.fetch` or to
      :class:`comms.eventstreams.EventStreams`. It can also differ if
      :func:`comms.http.fetch` is used with *use_fake_user_agent* set to
      ``True`` or to a custom UA string, or if
      *fake_user_agent_exceptions* is defined in the :mod:`config` file.
"""
#
# (C) Pywikibot team, 2007-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import os
import sys
from pathlib import Path

import pywikibot
from pywikibot.comms.http import user_agent
from pywikibot.version import getversion


class DummyModule:

    """Fake module instance."""

    __version__ = 'n/a'


try:
    import packaging
except ImportError:
    packaging = DummyModule()

try:
    import mwparserfromhell
except ImportError:
    mwparserfromhell = DummyModule()

try:
    import wikitextparser
except ImportError:
    wikitextparser = DummyModule()

try:
    import requests
except ImportError:
    requests = DummyModule()


WMF_CACERT = 'MIIDxTCCAq2gAwIBAgIQAqxcJmoLQJuPC3nyrkYldzANBgkqhkiG9w0BAQUFADBs'


def main(*args: str) -> None:
    """Print pywikibot version and important settings.

    .. versionchanged:: 9.1.2
       usernames are not printed with ``-nouser`` option.
    """
    pywikibot.info('Pywikibot: ' + getversion())
    pywikibot.info('Release version: ' + pywikibot.__version__)
    pywikibot.info('packaging version: ' + packaging.__version__)
    pywikibot.info('mwparserfromhell version: ' + mwparserfromhell.__version__)
    pywikibot.info('wikitextparser version: ' + wikitextparser.__version__)
    pywikibot.info('requests version: ' + requests.__version__)

    has_wikimedia_cert = False
    if (not hasattr(requests, 'certs')
            or not hasattr(requests.certs, 'where')
            or not callable(requests.certs.where)):
        pywikibot.info('  cacerts: not defined')
    else:
        cert = Path(requests.certs.where())
        # is_symlink() required for Python 3.12 and below.
        # Otherwise follow_symlinks=True could be used in is_file().
        if not cert.is_file() or cert.is_symlink():
            pywikibot.info(f'  cacerts: {cert.name} (missing)')
        else:
            pywikibot.info(f'  cacerts: {cert}')
            text = cert.read_text(encoding='utf-8')
            if WMF_CACERT in text:
                has_wikimedia_cert = True
            pywikibot.info('    certificate test: {}'
                           .format('ok' if has_wikimedia_cert else 'not ok'))

    if not has_wikimedia_cert:
        pywikibot.info('  Please reinstall requests!')

    pywikibot.info('Python: ' + sys.version)
    pywikibot.info('User-Agent: ' + user_agent(pywikibot.Site()))

    # check environment settings
    settings = {key for key in os.environ if key.startswith('PYWIKIBOT')}
    settings.update(['PYWIKIBOT_DIR', 'PYWIKIBOT_DIR_PWB',
                     'PYWIKIBOT_NO_USER_CONFIG'])
    for environ_name in sorted(settings):
        pywikibot.info(
            '{}: {}'.format(environ_name,
                            os.environ.get(environ_name, 'Not set') or "''"))

    pywikibot.info('Config base dir: ' + pywikibot.config.base_dir)

    if '-nouser' in sys.argv:
        usernames_items = {}
    else:
        usernames_items = pywikibot.config.usernames.items()
    for family, usernames in usernames_items:
        if not usernames:
            continue
        pywikibot.info(f"Usernames for family '{family}':")
        for lang, username in usernames.items():
            pywikibot.info(f'\t{lang}: {username}')


if __name__ == '__main__':
    main()



"""Family module for WOW wiki."""
#
# (C) Pywikibot team, 2009-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from pywikibot import family
from pywikibot.tools import classproperty, deprecated


class Family(family.FandomFamily):

    """Family class for WOW Wiki."""

    name = 'wowwiki'
    domain = 'wowwiki.fandom.com'

    codes = {
        'ar', 'cs', 'da', 'de', 'el', 'en', 'es', 'et', 'fa', 'fi', 'fr', 'he',
        'hu', 'it', 'ja', 'ko', 'nl', 'no', 'pl', 'pt', 'pt-br', 'ru', 'uk',
        'zh', 'zh-tw',
    }

    removed_wikis = ['is', 'hr', 'lt', 'lv', 'ro', 'sk', 'sr', 'sv', 'tr']

    code_aliases = {'nn': 'no'}

    @classproperty
    def langs(cls):
        """Property listing family languages."""
        cls.langs = super().langs
        # override deviations
        for i, lang in enumerate(['en', 'es', 'et', 'uk'], start=1):
            cls.langs[lang] = cls.domains[i]
        return cls.langs

    @classproperty
    def disambiguationTemplates(cls):  # noqa: N802
        """Property listing disambiguation templates."""
        cls.disambiguationTemplates = super().disambiguationTemplates
        cls.disambiguationTemplates['en'] = ['disambig', 'disambig/quest',
                                             'disambig/quest2',
                                             'disambig/achievement2']
        return cls.disambiguationTemplates

    @classproperty
    def disambcatname(cls):
        """Property listing disambiguation category name."""
        cls.disambcatname = super().disambcatname
        cls.disambcatname['en'] = 'Disambiguations'
        return cls.disambcatname

    # Wikia's default SelectCategory extension always puts categories last
    @classproperty
    @deprecated('site.has_extension("CategorySelect")', since='10.3.0')
    def categories_last(cls):
        """Property listing site keys for categories at last position.

        .. deprecated:: 10.3
           use :meth:`site.has_extension('CategorySelect')
           <pywikibot.site._apisite.APISite.has_extension>` instead
        """
        return cls.langs.keys()

    @classproperty
    def domains(cls):
        """List of domains used by family wowwiki."""
        return [cls.domain,
                'wowwiki-archive.fandom.com',  # en
                'wow.gamepedia.com',  # es
                'worldofwarcraft.fandom.com',  # et
                'warcraft.fandom.com']  # uk

    def scriptpath(self, code):
        """Return the script path for this family."""
        if code == 'es':
            return ''
        return super().scriptpath(code)



"""Decorators used by site models."""
#
# (C) Pywikibot team, 2008-2024
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import os
from textwrap import fill

from pywikibot.exceptions import UnknownExtensionError, UserRightsError
from pywikibot.tools import MediaWikiVersion, manage_wrapping


CLOSED_WIKI_MSG = (
    'Site {site} has been closed. Only steward can perform requested action.'
)


def must_be(group: str | None = None):
    """Decorator to require a certain user status when method is called.

    :param group: The group the logged in user should belong to. This
        parameter can be overridden by keyword argument 'as_group'.
    :return: method decorator
    :raises UserRightsError: user is not part of the required user
        group.
    """
    def decorator(fn):
        def callee(self, *args, **kwargs):
            grp = kwargs.pop('as_group', group)
            if self.obsolete:
                if not self.has_group('steward'):
                    raise UserRightsError(CLOSED_WIKI_MSG.format(site=self))

            elif not self.has_group(grp):
                raise UserRightsError(f'User "{self.user()}" is not part of '
                                      f'the required user group "{grp}"')

            return fn(self, *args, **kwargs)

        manage_wrapping(callee, fn)
        return callee

    return decorator


def need_extension(extension: str):
    """Decorator to require a certain MediaWiki extension.

    :param extension: the MediaWiki extension required
    :return: a decorator to make sure the requirement is satisfied when
        the decorated function is called.
    """
    def decorator(fn):
        def callee(self, *args, **kwargs):
            if not self.has_extension(extension):
                raise UnknownExtensionError(
                    f'Method "{fn.__name__}" is not implemented without the '
                    f'extension {extension}')
            return fn(self, *args, **kwargs)

        manage_wrapping(callee, fn)
        return callee

    return decorator


def need_right(right: str | None = None):
    """Decorator to require a certain user right when method is called.

    :param right: The right the logged in user should have.
    :return: method decorator
    :raises UserRightsError: user has insufficient rights.
    """
    def decorator(fn):
        def callee(self, *args, **kwargs):
            if self.obsolete:
                if not self.has_group('steward'):
                    raise UserRightsError(CLOSED_WIKI_MSG.format(site=self))

            elif right is not None and not self.has_right(right):
                if os.environ.get('PYWIKIBOT_TEST_RUNNING', '0') == '1':
                    rights = ' but:\n' + fill(
                        str(sorted(self.userinfo['rights'])),
                        width=76, break_on_hyphens=False)
                else:
                    rights = '.'
                raise UserRightsError(
                    f'User "{self.user()}" does not have required user right '
                    f'"{right}" on site {self}{rights}')
            return fn(self, *args, **kwargs)

        manage_wrapping(callee, fn)
        return callee

    return decorator


def need_version(version: str):
    """Decorator to require a certain MediaWiki version number.

    :param version: the mw version number required
    :return: a decorator to make sure the requirement is satisfied when
        the decorated function is called.
    """
    def decorator(fn):
        def callee(self, *args, **kwargs):
            if MediaWikiVersion(self.version()) < MediaWikiVersion(version):
                raise NotImplementedError(
                    f'Method or function "{fn.__name__}"\n'
                    f"isn't implemented in MediaWiki version < {version}")
            return fn(self, *args, **kwargs)

        manage_wrapping(callee, fn)

        return callee
    return decorator
